"""
Base state setter class.
"""
from abc import ABC, abstractmethod
from rlgym.utils.state_setters.wrappers.state_wrapper import StateWrapper


class StateSetter(ABC):
    @abstractmethod
    def reset(self, state_wrapper: StateWrapper):
        """
        Function to be called each time the environment is reset.

        :param state_wrapper: StateWrapper object to be modified with desired state values.

        NOTE: This function should change any desired values of the StateWrapper, which are all defaulted to 0.
        The values within StateWrapper are sent to the game each time the match is reset.
        """
        raise NotImplementedError



"""
The reward function.
"""

from abc import ABC, abstractmethod
from rlgym.utils.gamestates import GameState, PlayerData
import numpy as np


class RewardFunction(ABC):
    @abstractmethod
    def reset(self, initial_state: GameState):
        """
        Function to be called each time the environment is reset. This is meant to enable users to design stateful reward
        functions that maintain information about the game throughout an episode to determine a reward.

        :param initial_state: The initial state of the reset environment.
        """
        raise NotImplementedError

    @abstractmethod
    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        """
        Function to compute the reward for a player. This function is given a player argument, and it is expected that
        the reward returned by this function will be for that player.

        :param player: Player to compute the reward for.
        :param state: The current state of the game.
        :param previous_action: The action taken at the previous environment step.

        :return: A reward for the player provided.
        """
        raise NotImplementedError

    def get_final_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        """
        Function to compute the reward for a player at the final step of an episode. This will be called only once, when
        it is determined that the current state is a terminal one. This may be useful for sparse reward signals that only
        produce a value at the final step of an environment. By default, the regular get_reward is used.

        :param player: Player to compute the reward for.
        :param state: The current state of the game.
        :param previous_action: The action taken at the previous environment step.

        :return: A reward for the player provided.
        """
        return self.get_reward(player, state, previous_action)



from typing import List

import numpy as np
from rlgym.utils.gamestates import GameState, PlayerData
from rlgym.utils.reward_functions import DefaultReward


class MultiModelReward(DefaultReward):
    """
    Handles the distribution of rewards to specific models where each model uses a different reward function
    """
    index_tracker = 0

    def __init__(self, model_map: List[int], reward_funcs: List[DefaultReward]):
        """
        Handles the distribution of rewards to specific models where each model uses a different reward function

        :param model_map: A list containing the mapping of model index to player
        :param reward_funcs: A list of reward functions for each model, in the same order as the list
            of models used elsewhere
        """
        if max(model_map) >= len(reward_funcs):
            raise ValueError("model_map implies the existence of more models than reward funcs")

        super().__init__()
        self.model_map = model_map
        self.reward_funcs = reward_funcs
        # This will make sure the right instance index is passed
        self.index = self.index_tracker
        self.index_tracker += 1

    def reset(self, initial_state: GameState):
        for func in self.reward_funcs:
            func.reset(initial_state)

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        # look up which model it is
        model_num = self.model_map[
            self.index * len(state.players) + [bot.car_id for bot in state.players].index(player.car_id)]

        return self.reward_funcs[model_num].get_reward(player, state, previous_action)

    def get_final_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        # look up which model it is
        model_num = self.model_map[
            self.index * len(state.players) + [bot.car_id for bot in state.players].index(player.car_id)]

        return self.reward_funcs[model_num].get_final_reward(player, state, previous_action)


import numpy as np

from rlgym.utils import TerminalCondition, RewardFunction
from rlgym.utils.common_values import BALL_RADIUS
from rlgym.utils.gamestates import GameState, PlayerData
from rlgym.utils.reward_functions.common_rewards import ConstantReward
from rlgym.utils.state_setters import StateWrapper
from rlgym.utils.state_setters.random_state import RandomState, X_MAX, Y_MAX, PITCH_MAX, Z_MAX_BALL, YAW_MAX

MAX_CAR_HEIGHT = 62  # Slightly higher than Merc height at back
MIN_CAR_Z = 29  # Slightly below Batmobile z-coordinate (not top of car)


class BallOnGround(TerminalCondition):
    def reset(self, initial_state: GameState):
        pass

    def is_terminal(self, current_state: GameState) -> bool:
        # Check if bottom of ball is below car center of mass.
        # Technically this can be recovered from but there's no easy way to check for ball on ground.
        return current_state.ball.position[2] < BALL_RADIUS + MIN_CAR_Z


class DribbleReward(RewardFunction):
    def reset(self, initial_state: GameState):
        pass

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        return 0

    def get_final_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        return -1


class DribbleState(StateWrapper):
    def reset(self, state_wrapper: StateWrapper):
        car = state_wrapper.cars[0]
        car.set_pos(
            np.random.uniform(-X_MAX, X_MAX),
            np.random.uniform(-Y_MAX, Y_MAX),
            17,
        )
        car.set_rot(
            pitch=0,
            roll=0,
            yaw=np.random.uniform(-YAW_MAX, YAW_MAX)
        )
        r = np.random.normal(0, 10)
        theta = np.random.uniform(-np.pi, np.pi)
        eps_x, eps_y = r * np.cos(theta), r * np.sin(theta)
        state_wrapper.ball.set_pos(
            car.position[0] + eps_x,  # Displace a little bit so the ball is not balanced on the car
            car.position[1] + eps_y,
            MAX_CAR_HEIGHT + BALL_RADIUS,  # np.random.uniform(MAX_CAR_HEIGHT + BALL_RADIUS, Z_MAX_BALL)
        )
        theta = np.random.uniform(-np.pi, np.pi)
        eps_x, eps_y = r * np.cos(theta), r * np.sin(theta)
        state_wrapper.ball.set_lin_vel(
            x=eps_x,
            y=eps_y
        )


dribble_config = dict(
    reward_function=DribbleReward(),
    terminal_conditions=[BallOnGround()],
    state_setter=DribbleState(),
    team_size=1,
    spawn_opponents=False
)



"""
A class containing all data about a player in the game.
"""

from rlgym.utils.gamestates import PhysicsObject


class PlayerData(object):
    def __init__(self):
        self.car_id: int = -1
        self.team_num: int = -1
        self.match_goals: int = -1
        self.match_saves: int = -1
        self.match_shots: int = -1
        self.match_demolishes: int = -1
        self.boost_pickups: int = -1
        self.is_demoed: bool = False
        self.on_ground: bool = False
        self.ball_touched: bool = False
        self.has_flip: bool = False
        self.boost_amount: float = -1
        self.car_data: PhysicsObject = PhysicsObject()
        self.inverted_car_data: PhysicsObject = PhysicsObject()

    def __str__(self):
        output = "****PLAYER DATA OBJECT****\n" \
                 "Match Goals: {}\n" \
                 "Match Saves: {}\n" \
                 "Match Shots: {}\n" \
                 "Match Demolishes: {}\n" \
                 "Boost Pickups: {}\n" \
                 "Is Alive: {}\n" \
                 "On Ground: {}\n" \
                 "Ball Touched: {}\n" \
                 "Has Flip: {}\n" \
                 "Boost Amount: {}\n" \
                 "Car Data: {}\n" \
                 "Inverted Car Data: {}"\
            .format(self.match_goals,
                    self.match_saves,
                    self.match_shots,
                    self.match_demolishes,
                    self.boost_pickups,
                    self.is_demoed,
                    self.on_ground,
                    self.ball_touched,
                    self.has_flip,
                    self.boost_amount,
                    self.car_data,
                    self.inverted_car_data)
        return output


import sys
import traceback

UNKNOWN_ERROR = -1
EOF_ERROR = 0
BROKEN_PIPE_ERROR = 1


def _eof_error():
    print("RLGym Local Pipe has crashed! Please close Rocket League and re-launch RLGym.")
    return EOF_ERROR


def handle_exception(e):
    args = e.args
    exception_code = UNKNOWN_ERROR
    print("RLGym has encountered an exception!\nException traceback:\n{}".format(traceback.format_exc()))
    if "The pipe has been ended." in args:
        exception_code = _eof_error()
    else:
        print("EXCEPTION ARGS:", args)

    return exception_code


class CrashPolicy:
    ATTEMPT_RECOVER = 0
    RAISE = 1
    ATTEMPT_RECOVER_THEN_RAISE = 3



from rlgym.utils.gamestates import PhysicsObject
import numpy as np


class PhysicsWrapper(object):

    def __init__(self, phys_obj: PhysicsObject = None):
        """
        PhysicsWrapper constructor. Under most circumstances, users should not expect to instantiate their own PhysicsWrapper objects.

        :param phys_obj: PhysicsObject object from which values will be read.
        """
        if phys_obj is None:
            self.position: np.ndarray = np.asarray([0, 0, 93])
            self.linear_velocity: np.ndarray = np.zeros(3)
            self.angular_velocity: np.ndarray = np.zeros(3)
        else:
            self._read_from_physics_object(phys_obj)

    def _read_from_physics_object(self, phys_obj: PhysicsObject):
        """
        A function to modify PhysicsWrapper values from values in a PhysicsObject object.
        """
        self.position = phys_obj.position
        self.linear_velocity = phys_obj.linear_velocity
        self.angular_velocity = phys_obj.angular_velocity

    def set_pos(self, x: float = None, y: float = None, z: float = None):
        """
        Sets position.

        :param x: Float indicating x position value.
        :param y: Float indicating y position value.
        :param z: Float indicating z position value.
        """
        if x is not None:
            self.position[0] = x
        if y is not None:
            self.position[1] = y
        if z is not None:
            self.position[2] = z

    def set_lin_vel(self, x: float = None, y: float = None, z: float = None):
        """
        Sets linear velocity.

        :param x: Float indicating x velocity value.
        :param y: Float indicating y velocity value.
        :param z: Float indicating z velocity value.
        """
        if x is not None:
            self.linear_velocity[0] = x
        if y is not None:
            self.linear_velocity[1] = y
        if z is not None:
            self.linear_velocity[2] = z

    def set_ang_vel(self, x: float = None, y: float = None, z: float = None):
        """
        Sets angular velocity.

        :param x: Float indicating x angular velocity value.
        :param y: Float indicating y angular velocity value.
        :param z: Float indicating z angular velocity value.
        """
        if x is not None:
            self.angular_velocity[0] = x
        if y is not None:
            self.angular_velocity[1] = y
        if z is not None:
            self.angular_velocity[2] = z

    def _encode(self) -> list:
        """
        Function called by a StateWrapper to produce a state string.

        :return: String containing value data.
        """
        encoded = np.concatenate((self.position, self.linear_velocity, self.angular_velocity))
        return encoded.tolist()



import numpy as np
from stable_baselines3.common.vec_env import VecEnvWrapper, VecEnv
from stable_baselines3.common.vec_env.base_vec_env import VecEnvObs, VecEnvStepReturn

from rlgym.utils.common_values import BLUE_TEAM
from rlgym.utils.gamestates import GameState


class SB3DistributeRewardsWrapper(VecEnvWrapper):
    """
    Does the same thing as DistributeRewards, but as a SB3 wrapper to enable logging of rewards before distributing.
    """
    def __init__(self, venv: VecEnv, team_spirit=0.3):
        super().__init__(venv)
        self.team_spirit = team_spirit

    def reset(self) -> VecEnvObs:
        return self.venv.reset()

    def step_wait(self) -> VecEnvStepReturn:
        observations, rewards, dones, infos = self.venv.step_wait()
        last_state = None
        for i, info in enumerate(infos):
            state: GameState = info["state"]
            if state != last_state:
                blue_indices = []
                orange_indices = []
                for n, player in enumerate(state.players):
                    if player.team_num == BLUE_TEAM:
                        blue_indices.append(n + i)
                    else:
                        orange_indices.append(n + i)

                for team, opp in (blue_indices, orange_indices), (orange_indices, blue_indices):
                    rewards[team] = (1 - self.team_spirit) * rewards[team] \
                                    + self.team_spirit * rewards[team].mean() \
                                    - np.nan_to_num(rewards[opp].mean())  # In case of no opponent

                last_state = state

        return observations, rewards, dones, infos



import numpy as np

from rlgym.utils import RewardFunction, math
from rlgym.utils.common_values import BALL_RADIUS, CAR_MAX_SPEED
from rlgym.utils.gamestates import GameState, PlayerData


class LiuDistancePlayerToBallReward(RewardFunction):
    def reset(self, initial_state: GameState):
        pass

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        # Compensate for inside of ball being unreachable (keep max reward at 1)
        dist = np.linalg.norm(player.car_data.position - state.ball.position) - BALL_RADIUS
        return np.exp(-0.5 * dist / CAR_MAX_SPEED)  # Inspired by https://arxiv.org/abs/2105.12196


class VelocityPlayerToBallReward(RewardFunction):
    def __init__(self, use_scalar_projection=False):
        super().__init__()
        self.use_scalar_projection = use_scalar_projection

    def reset(self, initial_state: GameState):
        pass

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        vel = player.car_data.linear_velocity
        pos_diff = state.ball.position - player.car_data.position
        if self.use_scalar_projection:
            # Vector version of v=d/t <=> t=d/v <=> 1/t=v/d
            # Max value should be max_speed / ball_radius = 2300 / 92.75 = 24.8
            # Used to guide the agent towards the ball
            inv_t = math.scalar_projection(vel, pos_diff)
            return inv_t
        else:
            # Regular component velocity
            norm_pos_diff = pos_diff / np.linalg.norm(pos_diff)
            vel /= CAR_MAX_SPEED
            return float(np.dot(norm_pos_diff, vel))


class FaceBallReward(RewardFunction):
    def reset(self, initial_state: GameState):
        pass

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        pos_diff = state.ball.position - player.car_data.position
        norm_pos_diff = pos_diff / np.linalg.norm(pos_diff)
        return float(np.dot(player.car_data.forward(), norm_pos_diff))


class TouchBallReward(RewardFunction):
    def __init__(self, aerial_weight=0.):
        self.aerial_weight = aerial_weight

    def reset(self, initial_state: GameState):
        pass

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        if player.ball_touched:
            # Default just rewards 1, set aerial weight to reward more depending on ball height
            return ((state.ball.position[2] + BALL_RADIUS) / (2 * BALL_RADIUS)) ** self.aerial_weight
        return 0



import numpy as np
import gym.spaces
from rlgym.utils.gamestates import GameState
from rlgym.utils.action_parsers import ActionParser


class DiscreteAction(ActionParser):
    """
    Simple discrete action space. All the analog actions have 3 bins by default: -1, 0 and 1.
    """

    def __init__(self, n_bins=3):
        super().__init__()
        assert n_bins % 2 == 1, "n_bins must be an odd number"
        self._n_bins = n_bins

    def get_action_space(self) -> gym.spaces.Space:
        return gym.spaces.MultiDiscrete([self._n_bins] * 5 + [2] * 3)

    def parse_actions(self, actions: np.ndarray, state: GameState) -> np.ndarray:
        actions = actions.reshape((-1, 8))

        # map all ternary actions from {0, 1, 2} to {-1, 0, 1}.
        actions[..., :5] = actions[..., :5] / (self._n_bins // 2) - 1

        return actions



import random

import numpy as np
import torch
from gym.spaces import MultiDiscrete
from rlgym.envs import Match
from rlgym.utils.reward_functions import CombinedReward
from stable_baselines3 import PPO
from stable_baselines3.common.callbacks import CheckpointCallback
from stable_baselines3.common.vec_env import VecMonitor, VecNormalize, VecCheckNan, VecEnvWrapper, VecEnv
from stable_baselines3.common.vec_env.base_vec_env import VecEnvObs, VecEnvStepReturn
from stable_baselines3.ppo import MlpPolicy

from rlgym.utils.obs_builders import AdvancedObs
from rlgym.utils.reward_functions.common_rewards import VelocityPlayerToBallReward, LiuDistancePlayerToBallReward, \
    EventReward
from rlgym.utils.state_setters import DefaultState
from rlgym.utils.terminal_conditions.common_conditions import TimeoutCondition, GoalScoredCondition

from rlgym_tools.extra_rewards.diff_reward import DiffReward
from rlgym_tools.sb3_utils import SB3MultipleInstanceEnv
from rlgym_tools.sb3_utils.sb3_multidiscrete_wrapper import SB3MultiDiscreteWrapper


class SB3MultiDiscreteWrapperV2(VecEnvWrapper):
    """
    Simply converts env such that action space is MultiDiscrete instead of Box (basically KBM).
    """

    def __init__(self, venv: VecEnv, n=3):
        super().__init__(venv)
        assert n % 2 == 1
        self.action_space = MultiDiscrete((n, n, 2, 2, 2))
        self.n = n

    def reset(self) -> VecEnvObs:
        return self.venv.reset()

    def step_async(self, actions: np.ndarray) -> None:
        actions = np.array(actions, copy=True)
        actions = actions.reshape((-1, 5))
        actions[:, 0] = actions[:, 0] / (self.n // 2) - 1
        actions[:, 1] = actions[:, 1] / (self.n // 2) - 1

        parsed = np.zeros((actions.shape[0], 8))
        parsed[:, 0] = actions[:, 0]  # throttle
        parsed[:, 1] = actions[:, 1]  # steer
        parsed[:, 2] = actions[:, 0]  # pitch
        parsed[:, 3] = actions[:, 1] * (1 - actions[:, 4])  # yaw
        parsed[:, 4] = actions[:, 1] * actions[:, 4]  # roll
        parsed[:, 5] = actions[:, 2]  # jump
        parsed[:, 6] = actions[:, 3]  # boost
        parsed[:, 7] = actions[:, 4]  # handbrake
        self.venv.step_async(parsed)

    def step_wait(self) -> VecEnvStepReturn:
        return self.venv.step_wait()


if __name__ == '__main__':  # Required for multiprocessing
    random.seed(123)
    np.random.seed(123)
    torch.manual_seed(123)
    frame_skip = 8  # Number of ticks to repeat an action
    half_life_seconds = 5  # Easier to conceptualize, after this many seconds the reward discount is 0.5

    fps = 120 / frame_skip
    gamma = np.exp(np.log(0.5) / (fps * half_life_seconds))  # Quick mafs
    print(f"fps={fps}, gamma={gamma})")


    def get_match():  # Need to use a function so that each instance can call it and produce their own objects
        return Match(
            team_size=2,  # 3v3 to get as many agents going as possible, will make results more noisy
            tick_skip=frame_skip,
            reward_function=CombinedReward.from_zipped(
                DiffReward(LiuDistancePlayerToBallReward()),
                EventReward(touch=1.)
            ),  # Simple reward since example code
            self_play=True,
            terminal_conditions=[TimeoutCondition(round(fps * 30)), GoalScoredCondition()],  # Some basic terminals
            obs_builder=AdvancedObs(),  # Not that advanced, good default
            state_setter=DefaultState()  # Resets to kickoff position
        )


    env = SB3MultipleInstanceEnv(get_match, 6, wait_time=30)  # Start 2 instances, waiting 60 seconds between each
    env = SB3MultiDiscreteWrapper(env)  # Convert action space to multidiscrete
    env = VecCheckNan(env)  # Optional
    env = VecMonitor(env)  # Recommended, logs mean reward and ep_len to Tensorboard
    env = VecNormalize(env, norm_obs=False, gamma=gamma)  # Highly recommended, normalizes rewards

    # Hyperparameters presumably better than default; inspired by original PPO paper
    model = PPO(
        MlpPolicy,
        env,
        n_epochs=32,  # PPO calls for multiple epochs
        learning_rate=1e-5,  # Around this is fairly common for PPO
        ent_coef=0.01,  # From PPO Atari
        vf_coef=1.,  # From PPO Atari
        gamma=gamma,  # Gamma as calculated using half-life
        verbose=3,  # Print out all the info as we're going
        batch_size=4096,  # Batch size as high as possible within reason
        n_steps=4096,  # Number of steps to perform before optimizing network
        tensorboard_log="out/logs",  # `tensorboard --logdir out/logs` in terminal to see graphs
        device="auto"  # Uses GPU if available
    )

    # Save model every so often
    # Divide by num_envs (number of agents) because callback only increments every time all agents have taken a step
    # This saves to specified folder with a specified name
    # callback = CheckpointCallback(round(1_000_000 / env.num_envs), save_path="policy", name_prefix="rl_model")

    model.learn(100_000_000)

    # # Now, if one wants to load a trained model from a checkpoint, use this function
    # # This will contain all the attributes of the original model
    # # Any attribute can be overwritten by using the custom_objects parameter,
    # # which includes n_envs (number of agents), which has to be overwritten to use a different amount
    # model = PPO.load("policy/rl_model_1000002_steps.zip", env, custom_objects=dict(n_envs=env.num_envs))
    # env.reset()  # Important when loading models, SB3 does not do this for you
    # # Use reset_num_timesteps=False to keep going with same logger/checkpoints
    # model.learn(100_000_000, callback=callback, reset_num_timesteps=False)



from rlgym.utils.state_setters import StateSetter
from rlgym.utils.state_setters import StateWrapper
from rlgym.utils.math import rand_vec3
import numpy as np
from numpy import random as rand
import random

GOAL_X_MAX = 800.0
GOAL_X_MIN = -800.0

PLACEMENT_BOX_X = 5000
PLACEMENT_BOX_Y = 2000
PLACEMENT_BOX_Y_OFFSET = 3000

GOAL_LINE = 5100

YAW_MAX = np.pi


class GoaliePracticeState(StateSetter):

    def __init__(self, aerial_only=False, allow_enemy_interference=False, first_defender_in_goal=False,
                 reset_to_max_boost=True):
        """
        GoaliePracticeState constructor.

        :param aerial_only: Boolean indicating whether the shots will only be in the air.
        :param allow_enemy_interference: Boolean indicating whether opponents will spawn close enough to easily affect the play
        :param first_defender_in_goal: Boolean indicating whether the first defender will spawn in the goal
        :param reset_to_max_boost: Boolean indicating whether the cars will start each episode with 100 boost or keep from last episode
        """
        super().__init__()
        self.team_turn = 0  # swap every reset who's getting shot at

        self.aerial_only = aerial_only
        self.allow_enemy_interference = allow_enemy_interference
        self.first_defender_in_goal = first_defender_in_goal
        self.reset_to_max_boost = reset_to_max_boost

    def reset(self, state_wrapper: StateWrapper):
        """
        Modifies the StateWrapper to set a new shot

        :param state_wrapper: StateWrapper object to be modified with desired state values.
        """
        self._reset_ball(state_wrapper, self.team_turn, self.aerial_only)
        self._reset_cars(state_wrapper, self.team_turn, self.first_defender_in_goal, self.allow_enemy_interference,
                         self.reset_to_max_boost)

        # which team will recieve the next incoming shot
        self.team_turn = (self.team_turn + 1) % 2

    def _reset_cars(self, state_wrapper: StateWrapper, team_turn, first_defender_in_goal, allow_enemy_interference,
                    reset_to_max_boost):
        """
        Function to set cars in preparation for an incoming shot

        :param state_wrapper: StateWrapper object to be modified.
        :param team_turn: team who's getting shot at
        :param allow_enemy_interference: Boolean indicating whether opponents will spawn close enough to easily affect the play
        :param first_defender_in_goal: Boolean indicating whether the first defender will spawn in the goal
        :param reset_to_max_boost: Boolean indicating whether the cars will start each episode with 100 boost or keep from last episode
        """
        first_set = False
        for car in state_wrapper.cars:
            # set random position and rotation for all cars based on pre-determined ranges

            if car.team_num == team_turn and not first_set:
                if first_defender_in_goal:
                    y_pos = -GOAL_LINE if car.team_num == 0 else GOAL_LINE
                    car.set_pos(0, y_pos, z=17)
                    first_set = True
                else:
                    self._place_car_in_box_area(car, team_turn)

            else:
                if allow_enemy_interference:
                    self._place_car_in_box_area(car, team_turn)

                else:
                    self._place_car_in_box_area(car, car.team_num)

            if reset_to_max_boost:
                car.boost = 100

            car.set_rot(0, rand.random() * YAW_MAX - YAW_MAX / 2, 0)

    def _place_car_in_box_area(self, car, team_delin):
        """
        Function to place a car in an allowed areaI 

        :param car: car to be modified
        :param team_delin: team number delinator to look at when deciding where to place the car
        """

        y_pos = (PLACEMENT_BOX_Y - (rand.random() * PLACEMENT_BOX_Y))

        if team_delin == 0:
            y_pos -= PLACEMENT_BOX_Y_OFFSET
        else:
            y_pos += PLACEMENT_BOX_Y_OFFSET

        car.set_pos(rand.random() * PLACEMENT_BOX_X - PLACEMENT_BOX_X / 2, y_pos, z=17)

    def _reset_ball(self, state_wrapper: StateWrapper, team_turn, aerial_only):
        """
        Function to set a new ball towards a goal

        :param state_wrapper: StateWrapper object to be modified.
        :param team_turn: team who's getting shot at
        :param aerial_only: Boolean indicating whether should shots only be from the air
        """

        pos, lin_vel, ang_vel = self._get_shot_parameters(team_turn, aerial_only)
        state_wrapper.ball.set_pos(pos[0], pos[1], pos[2])
        state_wrapper.ball.set_lin_vel(lin_vel[0], lin_vel[1], lin_vel[2])
        state_wrapper.ball.set_ang_vel(ang_vel[0], ang_vel[1], ang_vel[2])

    def _get_shot_parameters(self, team_turn, aerial_only):
        """
        Function to set a new ball towards a goal
        
        :param team_turn: team who's getting shot at
        :param aerial_only: Boolean indicating whether should shots only be from the air
        """

        # *** Magic numbers are from manually calibrated shots ***
        # *** They are unrelated to numbers in other functions ***

        shotpick = random.randrange(4)
        INVERT_IF_BLUE = (-1 if team_turn == 0 else 1)  # invert shot for blue

        # random pick x value of target in goal
        x_pos = random.uniform(GOAL_X_MIN, GOAL_X_MAX)

        # if its not an air shot, we can randomize the shot speed
        shot_randomizer = 1 if aerial_only else (random.uniform(.2, 1))

        y_vel = (3000 * INVERT_IF_BLUE) if aerial_only else (3000 * shot_randomizer * INVERT_IF_BLUE)
        if shotpick == 0:  # long range shot

            z_pos = 1500 if aerial_only else random.uniform(100, 1500)

            pos = np.array([x_pos, -3300 * INVERT_IF_BLUE, z_pos])
            lin_vel = np.array([0, y_vel, 600])
        elif shotpick == 1:  # medium range shot
            z_pos = 1550 if aerial_only else random.uniform(100, 1550)

            pos = np.array([x_pos, -500 * INVERT_IF_BLUE, z_pos])
            lin_vel = np.array([0, y_vel, 100])

        elif shotpick == 2:  # angled shot
            z_pos = 1500 if aerial_only else random.uniform(100, 1500)
            x_pos += 3200  # add offset to start the shot from the side

            pos = np.array([x_pos, 0, z_pos])
            lin_vel = np.array([-1900 * shot_randomizer, y_vel, 0])

        elif shotpick == 3:  # opposite angled shot
            z_pos = 1500 if aerial_only else random.uniform(100, 1500)
            x_pos -= 3200  # add offset to start the shot from the other side

            pos = np.array([x_pos, 0, z_pos])
            lin_vel = np.array([1900 * shot_randomizer, y_vel, 0])
        else:
            print("FAULT")

        ang_vel = np.array([0, 0, 0])

        return pos, lin_vel, ang_vel



"""
A module containing implementations of common terminal conditions.
"""

from rlgym.utils.terminal_conditions import TerminalCondition
from rlgym.utils.gamestates import GameState


class TimeoutCondition(TerminalCondition):
    """
    A condition that will terminate an episode after some number of steps.
    """

    def __init__(self, max_steps: int):
        super().__init__()
        self.steps = 0
        self.max_steps = max_steps

    def reset(self, initial_state: GameState):
        """
        Reset the step counter.
        """

        self.steps = 0

    def is_terminal(self, current_state: GameState) -> bool:
        """
        Increment the current step counter and return `True` if `max_steps` have passed.
        """

        self.steps += 1
        return self.steps >= self.max_steps


class NoTouchTimeoutCondition(TimeoutCondition):
    def is_terminal(self, current_state: GameState):
        if any(p.ball_touched for p in current_state.players):
            self.steps = 0
            return False
        else:
            return super(NoTouchTimeoutCondition, self).is_terminal(current_state)


class GoalScoredCondition(TerminalCondition):
    """
    A condition that will terminate an episode as soon as a goal is scored by either side.
    """

    def __init__(self):
        super().__init__()
        self.blue_score = 0
        self.orange_score = 0

    def reset(self, initial_state: GameState):
        pass

    def is_terminal(self, current_state: GameState) -> bool:
        """
        Check to see if the game score for either team has been changed. If either score has changed, update the current
        known scores for both teams and return `True`. Note that the known game scores are never reset for this object
        because the game score is not set to 0 for both teams at the beginning of an episode.
        """

        if current_state.blue_score != self.blue_score or current_state.orange_score != self.orange_score:
            self.blue_score = current_state.blue_score
            self.orange_score = current_state.orange_score
            return True
        return False


class BallTouchedCondition(TerminalCondition):

    def __init__(self):
        super().__init__()
        self.last_touch = None

    def reset(self, initial_state: GameState):
        self.last_touch = initial_state.last_touch

    def is_terminal(self, current_state: GameState) -> bool:
        """
        Return `True` if the last touch does not have the same ID as the last touch from the initial state.
        """
        return current_state.last_touch != self.last_touch



import json
import os
import subprocess
import webbrowser
from pathlib import Path
from time import sleep
import psutil
from typing import Set, Tuple, Union, List, Optional
import re

# Copied from https://github.com/RLBot/RLBot/blob/master/src/main/python/rlbot/gamelaunch/epic_launch.py

def launch_with_epic_simple(ideal_args: List[str]) -> Optional[subprocess.Popen]:
    try:
        # Try launch via Epic Games
        epic_rl_exe_path = locate_epic_games_launcher_rocket_league_binary()
        if epic_rl_exe_path is not None:
            exe_and_args = [str(epic_rl_exe_path)] + ideal_args + ['-EpicPortal']
            # print(f'Launching Rocket League with: {exe_and_args}')
            try:
                return subprocess.Popen(exe_and_args)
            except Exception as e:
                print(f'Unable to launch via Epic due to: {e}')
    except:
        print('Unable to launch via Epic.')


def launch_with_epic_login_trick(ideal_args: List[str]) -> Optional[subprocess.Popen]:
    process = get_running_process('RocketLeague.exe', {'-pipe'})
    if process is None:
        args = get_epic_login_trick_args(ideal_args)
        if args is None:
            return
    else:
        args = process.cmdline()
        # Replace pipe_id with new one
        args[2] = ideal_args[1]

    return subprocess.Popen(args)

def get_epic_login_trick_args(ideal_args: List[str]) -> Optional[subprocess.Popen]:
    try:
        webbrowser.open('com.epicgames.launcher://apps/Sugar?action=launch&silent=true')
        process = None
        for i in range(10):
            sleep(1)
            process = get_running_process('RocketLeague.exe', {'-EpicPortal'})
            if process:
                break

        if process is None:
            return

        # get the args from the process
        all_args = process.cmdline()

        process.kill()
        all_args[1:1] = ideal_args
        return all_args
    except:
        return


def locate_epic_games_launcher_rocket_league_binary() -> Optional[Path]:
    # Make sure we're on windows, this will go poorly otherwise
    try:
        import winreg
    except ImportError:
        return

    # List taken from https://docs.unrealengine.com/en-US/GettingStarted/Installation/MultipleLauncherInstalls/index.html
    possible_registry_locations = (
        (winreg.HKEY_LOCAL_MACHINE, 'SOFTWARE\\Epic Games\\EpicGamesLauncher'),
        (winreg.HKEY_LOCAL_MACHINE, 'SOFTWARE\\WOW6432Node\\Epic Games\\EpicGamesLauncher'),
        (winreg.HKEY_CURRENT_USER, 'SOFTWARE\\Epic Games\\EpicGamesLauncher'),
        (winreg.HKEY_CURRENT_USER, 'SOFTWARE\\WOW6432Node\\Epic Games\\EpicGamesLauncher')
    )

    def search_for_manifest_file(app_data_path: Path) -> Optional[Path]:
        # Loop through the files ending in *.item in app_data_path/Manifests
        # Parse them as JSON and locate the one where MandatoryAppFolderName is 'rocketleague'
        # Extract the binary location and return it.
        for file in app_data_path.glob("*.item"):
            with open(app_data_path / file, 'r') as f:
                try:
                    data = json.load(f)
                except Exception:
                    continue

            if data.get('MandatoryAppFolderName') == 'rocketleague':
                return data

    for possible_location in possible_registry_locations:
        try:
            # get the path to the launcher's game data stuff
            path = Path(winreg.QueryValueEx(winreg.OpenKey(possible_location[0], possible_location[1]), "AppDataPath")[0]) / 'Manifests'
        except Exception:
            # the path, or the key, might not exist
            # in this case, we'll just skip over it
            continue

        binary_data = search_for_manifest_file(path)

        if binary_data is not None:
            return Path(binary_data['InstallLocation']) / binary_data['LaunchExecutable']

    # Nothing found in registry? Try C:\ProgramData\Epic\EpicGamesLauncher
    # Or consider using %programdata%
    path = Path(os.getenv("programdata")) / "Epic" / "EpicGamesLauncher" / "Data" / "Manifests"

    if os.path.isdir(path):
        binary_data = search_for_manifest_file(path)

        if binary_data is not None:
            return Path(binary_data['InstallLocation']) / binary_data['LaunchExecutable']


def get_running_process(process_name, required_args: Set[str]) -> Optional[psutil.Process]:
    # Find processes which contain the program or script name.
    matching_processes = []
    for process in psutil.process_iter():
        try:
            p = process.name()
            if process_name in p:
                matching_processes.append(process)
        except psutil.NoSuchProcess:
            continue
    # If matching processes were found, check for correct arguments.
    if len(matching_processes) != 0:
        for process in matching_processes:
            try:
                args = process.cmdline()[1:]
                for required_arg in required_args:
                    matching_args = [arg for arg in args if re.match(required_arg, arg, flags=re.IGNORECASE) is not None]
                    # Skip this process because it does not have a matching required argument.
                    if len(matching_args) == 0:
                        break
                else:
                    # If this process has not been skipped, it matches all arguments.
                    return process
            except psutil.AccessDenied:
                print(f"Access denied when trying to look at cmdline of {process}!")
        # If we didn't return yet it means all matching programs were skipped.
        raise WrongProcessArgs(f"{process_name} is not running with required arguments: {required_args}!")
    # No matching processes.
    return None


class WrongProcessArgs(UserWarning):
    pass



"""
The action parser.
"""

from abc import ABC, abstractmethod
from rlgym.utils.gamestates import PlayerData, GameState
import gym.spaces
import numpy as np
from typing import List, Union, Tuple, Dict, Any


class ActionParser(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def get_action_space(self) -> gym.spaces.Space:
        """
        Function that returns the action space type. It will be called during the initialization of the environment.
        
        :return: The type of the action space
        """
        raise NotImplementedError

    @abstractmethod
    def parse_actions(self, actions: Any, state: GameState) -> np.ndarray:
        """
        Function that parses actions from the action space into a format that rlgym understands.
        The expected return value is a numpy float array of size (n, 8) where n is the number of agents.
        The second dimension is indexed as follows: throttle, steer, yaw, pitch, roll, jump, boost, handbrake.
        The first five values are expected to be in the range [-1, 1], while the last three values should be either 0 or 1.

        :param actions: An object of actions, as passed to the `env.step` function.
        :param state: The GameState object of the current state that were used to generate the actions.

        :return: the parsed actions in the rlgym format.
        """
        raise NotImplementedError



from typing import Any

import gym
import numpy as np

from rlgym.utils.action_parsers import DiscreteAction
from rlgym.utils.gamestates import GameState


class KBMAction(DiscreteAction):
    def get_action_space(self) -> gym.spaces.Space:
        # throttle/pitch, steer/pitch/roll, jump, boost and handbrake/air roll
        return gym.spaces.MultiDiscrete([self._n_bins] * 2 + [2] * 3)

    def parse_actions(self, actions: Any, state: GameState) -> np.ndarray:
        actions = actions.reshape((-1, 5))
        actions[:, 0] = actions[:, 0] / (self._n_bins // 2) - 1
        actions[:, 1] = actions[:, 1] / (self._n_bins // 2) - 1

        parsed = np.zeros((actions.shape[0], 8))
        parsed[:, 0] = actions[:, 0]  # throttle
        parsed[:, 1] = actions[:, 1]  # steer
        parsed[:, 2] = actions[:, 0]  # pitch
        parsed[:, 3] = actions[:, 1] * (1 - actions[:, 4])  # yaw
        parsed[:, 4] = actions[:, 1] * actions[:, 4]  # roll
        parsed[:, 5] = actions[:, 2]  # jump
        parsed[:, 6] = actions[:, 3]  # boost
        parsed[:, 7] = actions[:, 4]  # handbrake

        return parsed



import os
from typing import List

from rlgym.envs import Match
from rlgym.gamelaunch import LaunchPreference
from rlgym.utils.terminal_conditions import common_conditions
from rlgym.utils.reward_functions import DefaultReward
from rlgym.utils.obs_builders import DefaultObs
from rlgym.utils.action_parsers import DefaultAction
from rlgym.utils.state_setters import DefaultState


def make(game_speed: int = 100,
         tick_skip: int = 8,
         spawn_opponents: bool = False,
         self_play: bool = False,
         team_size: int = 1,
         terminal_conditions: List[object] = (common_conditions.TimeoutCondition(225), common_conditions.GoalScoredCondition()),
         reward_fn: object = DefaultReward(),
         obs_builder: object = DefaultObs(),
         action_parser: object = DefaultAction(),
         state_setter: object = DefaultState(),
         launch_preference: str = LaunchPreference.EPIC,
         use_injector: bool = False,
         force_paging: bool = False):
    """
    :param game_speed: The speed the physics will run at, leave it at 100 unless your game can't run at over 240fps
    :param tick_skip: The amount of physics ticks your action will be repeated for
    :param spawn_opponents: Whether you want opponents or not
    :param self_play: If there are agent controller opponents or not
    :param random_resets: If enabled cars and ball will spawn in random locations after every reset
    :param team_size: Cars per team
    :param terminal_conditions: List of terminal condition objects (rlgym.utils.TerminalCondition)
    :param reward_fn: Reward function object (rlgym.utils.RewardFunction)
    :param obs_builder: Observation builder object (rlgym.utils.ObsBuilder)
    :param action_parser: Action parser object (rlgym.utils.ActionParser)
    :param state_setter: State Setter object (rlgym.utils.StateSetter)
    :param launch_preference: Rocket League launch preference (rlgym.gamelaunch.LaunchPreference) or path to RocketLeague executable
    :param use_injector: Whether to use RLGym's bakkesmod injector or not. Enable if launching multiple instances
    :param force_paging: Enable forced paging of each spawned rocket league instance to reduce memory utilization
                             immediately, instead of allowing the OS to slowly page untouched allocations.
                             WARNING: This will require you to potentially expand your Windows Page File [1]
                             and it may substantially increase disk activity, leading to decreased disk lifetime.
                             Use at your own peril.
                             Default is off: OS dictates the behavior.
    :return: Gym object
    [1]: https://www.tomshardware.com/news/how-to-manage-virtual-memory-pagefile-windows-10,36929.html
    """

    # Imports are inside the function because setup fails otherwise (Missing win32file)
    from rlgym.gym import Gym
    from rlgym.version import print_current_release_notes

    print_current_release_notes()

    match = Match(reward_function=reward_fn,
                  terminal_conditions=terminal_conditions,
                  obs_builder=obs_builder,
                  action_parser=action_parser,
                  state_setter=state_setter,
                  team_size=team_size,
                  tick_skip=tick_skip,
                  game_speed=game_speed,
                  spawn_opponents=spawn_opponents,
                  self_play=self_play)

    return Gym(match, pipe_id=os.getpid(), launch_preference=launch_preference, use_injector=use_injector, force_paging=force_paging)



import numpy as np

from rlgym.utils import math
from rlgym.utils.common_values import BLUE_TEAM, BLUE_GOAL_BACK, ORANGE_GOAL_BACK, ORANGE_TEAM, BALL_MAX_SPEED, \
    CAR_MAX_SPEED
from rlgym.utils.gamestates import GameState, PlayerData
from rlgym.utils.reward_functions import RewardFunction


class EventReward(RewardFunction):
    def __init__(self, goal=0., team_goal=0., concede=-0., touch=0., shot=0., save=0., demo=0.):
        """
        :param goal: reward for goal scored by player.
        :param team_goal: reward for goal scored by player's team.
        :param concede: reward for goal scored by opponents. Should be negative if used as punishment.
        :param touch: reward for touching the ball.
        :param shot: reward for shooting the ball (as detected by Rocket League).
        :param save: reward for saving the ball (as detected by Rocket League).
        :param demo: reward for demolishing a player.
        """
        super().__init__()
        self.weights = np.array([goal, team_goal, concede, touch, shot, save, demo])

        # Need to keep track of last registered value to detect changes
        self.last_registered_values = {}

    @staticmethod
    def _extract_values(player: PlayerData, state: GameState):
        if player.team_num == BLUE_TEAM:
            team, opponent = state.blue_score, state.orange_score
        else:
            team, opponent = state.orange_score, state.blue_score

        return np.array([player.match_goals, team, opponent, player.ball_touched, player.match_shots,
                         player.match_saves, player.match_demolishes])

    def reset(self, initial_state: GameState, optional_data=None):
        # Update every reset since rocket league may crash and be restarted with clean values
        self.last_registered_values = {}
        for player in initial_state.players:
            self.last_registered_values[player.car_id] = self._extract_values(player, initial_state)

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray, optional_data=None):
        old_values = self.last_registered_values[player.car_id]
        new_values = self._extract_values(player, state)

        diff_values = new_values - old_values
        diff_values[diff_values < 0] = 0  # We only care about increasing values

        reward = np.dot(self.weights, diff_values)

        self.last_registered_values[player.car_id] = new_values
        return reward


class VelocityReward(RewardFunction):
    # Simple reward function to ensure the model is training.
    def __init__(self, negative=False):
        super().__init__()
        self.negative = negative

    def reset(self, initial_state: GameState):
        pass

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        return np.linalg.norm(player.car_data.linear_velocity) / CAR_MAX_SPEED * (1 - 2 * self.negative)


class SaveBoostReward(RewardFunction):
    def reset(self, initial_state: GameState):
        pass

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        # 1 reward for each frame with 100 boost, sqrt because 0->20 makes bigger difference than 80->100
        return np.sqrt(player.boost_amount)


class ConstantReward(RewardFunction):
    def reset(self, initial_state: GameState):
        pass

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        return 1


class AlignBallGoal(RewardFunction):
    def __init__(self, defense=1., offense=1.):
        super().__init__()
        self.defense = defense
        self.offense = offense

    def reset(self, initial_state: GameState):
        pass

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        ball = state.ball.position
        pos = player.car_data.position
        protecc = np.array(BLUE_GOAL_BACK)
        attacc = np.array(ORANGE_GOAL_BACK)
        if player.team_num == ORANGE_TEAM:
            protecc, attacc = attacc, protecc

        # Align player->ball and net->player vectors
        defensive_reward = self.defense * math.cosine_similarity(ball - pos, pos - protecc)

        # Align player->ball and player->net vectors
        offensive_reward = self.offense * math.cosine_similarity(ball - pos, attacc - pos)

        return defensive_reward + offensive_reward



"""
A class to represent the state of a physics object from the game.
"""

from rlgym.utils import math
import numpy as np
from typing import Optional


class PhysicsObject(object):
    def __init__(self, position=None, quaternion=None, linear_velocity=None, angular_velocity=None):
        self.position: np.ndarray = position if position is not None else np.zeros(3)

        # ones by default to prevent mathematical errors when converting quat to rot matrix on empty physics state
        self.quaternion: np.ndarray = quaternion if quaternion is not None else np.ones(4)

        self.linear_velocity: np.ndarray = linear_velocity if linear_velocity is not None else np.zeros(3)
        self.angular_velocity: np.ndarray = angular_velocity if angular_velocity is not None else np.zeros(3)
        self._euler_angles: Optional[np.ndarray] = np.zeros(3)
        self._rotation_mtx: Optional[np.ndarray] = np.zeros((3,3))
        self._has_computed_rot_mtx = False
        self._has_computed_euler_angles = False

    def decode_car_data(self, car_data: np.ndarray):
        """
        Function to decode the physics state of a car from the game state array.
        :param car_data: Slice of game state array containing the car data to decode.
        """
        self.position = car_data[:3]
        self.quaternion = car_data[3:7]
        self.linear_velocity = car_data[7:10]
        self.angular_velocity = car_data[10:]

    def decode_ball_data(self, ball_data: np.ndarray):
        """
        Function to decode the physics state of the ball from the game state array.
        :param ball_data: Slice of game state array containing the ball data to decode.
        """
        self.position = ball_data[:3]
        self.linear_velocity = ball_data[3:6]
        self.angular_velocity = ball_data[6:9]

    def forward(self) -> np.ndarray:
        return self.rotation_mtx()[:, 0]

    def right(self) -> np.ndarray:
        return self.rotation_mtx()[:, 1]

    def left(self) -> np.ndarray:
        return self.rotation_mtx()[:, 1] * -1

    def up(self) -> np.ndarray:
        return self.rotation_mtx()[:, 2]

    def pitch(self) -> float:
        return self.euler_angles()[0]

    def yaw(self) -> float:
        return self.euler_angles()[1]

    def roll(self) -> float:
        return self.euler_angles()[2]

    # pitch, yaw, roll
    def euler_angles(self) -> np.ndarray:
        if not self._has_computed_euler_angles:
            self._euler_angles = math.quat_to_euler(self.quaternion)
            self._has_computed_euler_angles = True

        return self._euler_angles

    def rotation_mtx(self) -> np.ndarray:
        if not self._has_computed_rot_mtx:
            self._rotation_mtx = math.quat_to_rot_mtx(self.quaternion)
            self._has_computed_rot_mtx = True

        return self._rotation_mtx

    def serialize(self):
        """
        Function to serialize all the values contained by this physics object into a single 1D list. This can be useful
        when constructing observations for a policy.
        :return: List containing the serialized data.
        """
        repr = []

        if self.position is not None:
            for arg in self.position:
                repr.append(arg)
                
        if self.quaternion is not None:
            for arg in self.quaternion:
                repr.append(arg)
                
        if self.linear_velocity is not None:
            for arg in self.linear_velocity:
                repr.append(arg)
                
        if self.angular_velocity is not None:
            for arg in self.angular_velocity:
                repr.append(arg)

        if self._euler_angles is not None:
            for arg in self._euler_angles:
                repr.append(arg)

        if self._rotation_mtx is not None:
            for arg in self._rotation_mtx.ravel():
                repr.append(arg)

        return repr



from collections import deque
from typing import Any

import numpy as np

from rlgym.utils.gamestates import PlayerData, GameState
from rlgym.utils.obs_builders import ObsBuilder


class GeneralStacker(ObsBuilder):
    """
    Stacks several observations into one

    :param obs: observation that is to be stacked.
    :param stack_size: how many past observations the queue hold.

    """

    def __init__(self, obs: ObsBuilder, stack_size: int = 15):
        super().__init__()
        self.stack_size = stack_size
        self.obs = obs
        self.queue = deque([], maxlen=self.stack_size)
        self.frame_n = 0
        self.new = True

    def reset(self, initial_state: GameState):
        self.new = True

    def build_obs(
            self, player: PlayerData, state: GameState, previous_action: np.ndarray
    ) -> Any:
        # sort out the edge case after reset of environment
        if self.new:
            initial_obs = self.obs.build_obs(player, state, previous_action)
            for _ in range(self.stack_size):
                self.queue.appendleft(initial_obs)
        self.new = False

        frame = self.obs.build_obs(player, state, previous_action)

        self.frame_n += 1
        self.queue.appendleft(frame)
        return np.concatenate(self.queue)



import json
import os
import subprocess
import sys
import time
from typing import Tuple, Iterator

import ballchasing
import numpy as np
import pandas as pd
import torch
from torch.nn import CrossEntropyLoss
from torch.optim import Adam
from torch.utils.data.dataset import T_co, IterableDataset
from tqdm.contrib.concurrent import process_map

from earl_pytorch import EARL
from earl_pytorch.dataset.create_dataset import get_base_features, normalize, swap_teams, swap_left_right
from earl_pytorch.util.constants import POS_X, BALL_COLS, DEFAULT_FEATURES_STR, PLAYER_COLS, IS_BLUE, IS_ORANGE, POS_Y, \
    POS_Z, BOOST_AMOUNT, IS_DEMOED, CLS
from earl_pytorch.util.util import rotator_to_matrix, NGPModel
from rlgym.utils.common_values import BOOST_LOCATIONS
from rlgym.utils.gamestates import GameState

command = r'carball.exe -i "{}" -o "{}" parquet'

ENV = os.environ.copy()
ENV["NO_COLOR"] = "1"

boost_locations = np.array(BOOST_LOCATIONS)


def euler_to_quaternion(pitch, yaw, roll):
    cy = np.cos(yaw * 0.5)
    sy = np.sin(yaw * 0.5)
    cp = np.cos(pitch * 0.5)
    sp = np.sin(pitch * 0.5)
    cr = np.cos(roll * 0.5)
    sr = np.sin(roll * 0.5)

    w = cr * cp * cy + sr * sp * sy
    x = sr * cp * cy - cr * sp * sy
    y = cr * sp * cy + sr * cp * sy
    z = cr * cp * sy - sr * sp * cy

    return w, x, y, z


def process_replay(replay_path, output_folder):
    folder, fn = os.path.split(replay_path)
    replay_name = fn.replace(".replay", "")
    processed_folder = os.path.join(output_folder, replay_name)
    os.makedirs(processed_folder, exist_ok=True)

    with open(os.path.join(processed_folder, "carball.o.log"), "w", encoding="utf8") as stdout_f:
        with open(os.path.join(processed_folder, "carball.e.log"), "w", encoding="utf8") as stderr_f:
            return subprocess.run(
                command.format(replay_path, processed_folder),
                stdout=stdout_f,
                stderr=stderr_f,
                env=ENV
            )


class CarballAnalysis:
    METADATA_FNAME = "metadata.json"
    ANALYZER_FNAME = "analyzer.json"
    BALL_FNAME = "__ball.parquet"
    GAME_FNAME = "__game.parquet"
    PLAYER_FNAME = "player_{}.parquet"

    def __init__(self, processed_folder: str):
        # print(processed_folder, self.METADATA_FNAME)
        self.metadata = json.load(open(os.path.join(processed_folder, self.METADATA_FNAME)))
        self.analyzer = json.load(open(os.path.join(processed_folder, self.ANALYZER_FNAME)))

        self.ball = pd.read_parquet(os.path.join(processed_folder, self.BALL_FNAME))
        self.game = pd.read_parquet(os.path.join(processed_folder, self.GAME_FNAME))
        self.players = {}
        for player in self.metadata["players"]:
            uid = player["unique_id"]
            player_path = os.path.join(processed_folder, self.PLAYER_FNAME.format(uid))
            if os.path.exists(player_path):
                self.players[uid] = pd.read_parquet(player_path)

    def to_rlgym(self) -> Iterator[Tuple[GameState, np.ndarray]]:
        ball = self.ball.copy()
        # ball[["quat_w", "quat_x", "quat_y", "quat_z"]] = \
        #     np.asarray(euler_to_quaternion(ball["rot_pitch"], ball["rot_yaw"], ball["rot_roll"])).T
        game = self.game.copy()
        players = {}
        for uid, player_df in self.players.items():
            player_df = player_df.copy()
            # player_df[["quat_w", "quat_x", "quat_y", "quat_z"]] = \
            #     np.asarray(euler_to_quaternion(player_df["rot_pitch"], player_df["rot_yaw"], player_df["rot_roll"])).T
            players[uid] = player_df

        df = pd.DataFrame(index=game.index)
        df.loc[:, "ticks_since_last_transmit"] = (game["delta"].values * 120).round()
        df.loc[:, ["blue_score", "orange_score"]] = 0
        df.loc[:, [f"pad_{n}" for n in range(34)]] = 0

        physics_cols = ["pos_x", "pos_y", "pos_z",
                        "quat_w", "quat_x", "quat_y", "quat_z",
                        "vel_x", "vel_y", "vel_z",
                        "ang_vel_x", "ang_vel_y", "ang_vel_z"]
        invert = np.array([1, -1, 1,
                           1, -1, 1,
                           1, -1, 1])
        ball_physics_cols = [col for col in physics_cols if not col.startswith("quat")]
        df.loc[:, [f"ball/{col}" for col in ball_physics_cols]] = ball[ball_physics_cols].values
        df.loc[:, [f"inverted_ball/{col}" for col in ball_physics_cols]] = ball[ball_physics_cols].values * invert

        controls_df = pd.DataFrame(index=df.index)
        for uid, player_df in players.items():
            df.loc[:, f"{uid}/car_id"] = int(uid)
            df.loc[:, f"{uid}/team_num"] = next(
                int(p["is_orange"]) for p in self.metadata["players"] if p["unique_id"] == uid)
            df.loc[:, [f"{uid}/{col}" for col in physics_cols]] = player_df[physics_cols].values
            df.loc[:, [f"inverted_{uid}/{col}" for col in physics_cols]] = 0  # Get columns in right order first
            df.loc[:, [f"inverted_{uid}/{col}" for col in ball_physics_cols]] = player_df[ball_physics_cols].values \
                                                                                * invert
            df[[f"inverted_{uid}/quat_{axis}" for axis in "wxyz"]] = np.asarray(
                euler_to_quaternion(player_df["rot_pitch"], -player_df["rot_yaw"], player_df["rot_roll"])).T

            df.loc[:, f"{uid}/match_goals"] = player_df["match_goals"]
            df.loc[:, f"{uid}/match_saves"] = player_df["match_saves"]
            df.loc[:, f"{uid}/match_shots"] = player_df["match_shots"]
            df.loc[:, f"{uid}/match_demos"] = 0
            df.loc[:, f"{uid}/match_pickups"] = 0
            df.loc[:, f"{uid}/is_demoed"] = 0
            df.loc[:, f"{uid}/on_ground"] = False
            df.loc[:, f"{uid}/ball_touched"] = False
            df.loc[:, f"{uid}/has_flip"] = False
            df.loc[:, f"{uid}/boost_amount"] = player_df["boost_amount"] / 100

            controls_df.loc[:, f"{uid}/throttle"] = player_df["throttle"] / 127.5 - 1
            controls_df.loc[:, f"{uid}/steer"] = player_df["steer"] / 127.5 - 1
            controls_df.loc[:, f"{uid}/yaw"] = 0
            controls_df.loc[:, f"{uid}/pitch"] = 0
            controls_df.loc[:, f"{uid}/roll"] = 0
            controls_df.loc[:, f"{uid}/jump"] = 0
            controls_df.loc[:, f"{uid}/boost"] = player_df["boost_is_active"]
            controls_df.loc[:, f"{uid}/handbrake"] = player_df["handbrake"]

            for frame, pos in player_df[player_df["boost_pickup"] > 0][["pos_x", "pos_y", "pos_z"]].iterrows():
                boost_id = np.linalg.norm(boost_locations - pos.values, axis=-1).argmin()
                if boost_locations[boost_id][2] > 72:  # Big boost
                    df.loc[frame: frame + 30 * 10, f"pad_{boost_id}"] = 1
                else:  # Small boost
                    df.loc[frame: frame + 30 * 4, f"pad_{boost_id}"] = 1
                df.loc[frame:, f"{uid}/match_pickups"] += 1

        for goal in self.metadata["game"]["goals"]:
            if goal["is_orange"]:
                df.loc[goal["frame"]:, "orange_score"] += 1
            else:
                df.loc[goal["frame"]:, "blue_score"] += 1
            player_id = next(p["unique_id"] for p in self.metadata["players"] if p["name"] == goal["player_name"])
            df.loc[goal["frame"]:, f"{player_id}/match_goals"] += 1

        for demo in self.metadata["demos"]:
            frame = demo["frame_number"]
            attacker = demo["attacker_unique_id"]
            victim = demo["victim_unique_id"]
            df.loc[frame:, f"{attacker}/match_demos"] += 1
            df.loc[frame: frame + 30 * 3, f"{victim}/is_demoed"] = 1

        for hit in self.analyzer["hits"]:
            frame = hit["frame_number"]
            player_id = hit["player_unique_id"]
            df.loc[frame, f"{player_id}/ball_touched"] = 1

        for gameplay_period in self.analyzer["gameplay_periods"]:
            start_frame = gameplay_period["start_frame"]
            end_frame = gameplay_period["goal_frame"]
            if end_frame is None:
                end_frame = df.index[-2]

            yield from zip(
                df.loc[start_frame:end_frame].astype(float).fillna(0).apply(lambda x: GameState(list(x)), axis=1),
                controls_df[start_frame + 1:end_frame + 1]  # Actions are taken at t+1
                    .astype(float).fillna(0).apply(lambda x: np.reshape(x.values, (-1, 8)), axis=1)
            )

    def get_earl_compatible(self, n_players=None):
        if n_players is None:
            n_players = len(self.players)
        x, y = get_base_features(len(self.game), n_players)
        x_ball, x_boost, x_players = x
        y_score, y_next_touch, y_collect, y_demo, \
        y_throttle, y_steer, y_pitch, y_yaw, y_roll, y_jump, y_boost, y_handbrake = y

        player_team = {
            uid: next(int(p["is_orange"]) for p in self.metadata["players"] if p["unique_id"] == uid)
            for uid in self.players
        }

        x_boost[:, :, (POS_X, POS_Y, POS_Z)] = boost_locations
        x_boost[:, :, BOOST_AMOUNT] = 100

        valid_frames = []
        for gameplay_period in self.analyzer["gameplay_periods"]:
            start_frame = gameplay_period["start_frame"]
            end_frame = gameplay_period["end_frame"]
            goal_frame = gameplay_period["goal_frame"]
            if goal_frame is not None:
                end_frame = goal_frame  # Skip goal explosion
                goal = next(g for g in self.metadata["game"]["goals"] if g["frame"] == goal_frame)
                y_score[start_frame: end_frame] = int(goal["is_orange"])
                valid_frames.append(slice(start_frame, end_frame))

            ball_df = self.ball.iloc[start_frame: end_frame]
            x_ball[start_frame:end_frame, 0, BALL_COLS] = ball_df[[DEFAULT_FEATURES_STR[c] for c in BALL_COLS]]

            for p, (uid, player_df) in enumerate(self.players.items()):
                player_df = player_df.iloc[start_frame: end_frame].copy()
                forward, up = rotator_to_matrix(player_df["rot_yaw"], player_df["rot_pitch"], player_df["rot_roll"])
                player_df[['forward_x', 'forward_y', 'forward_z']] = np.array(forward).T
                player_df[['up_x', 'up_y', 'up_z']] = np.array(forward).T
                if player_team[uid] == 0:
                    x_players[start_frame:end_frame, p, IS_BLUE] = 1
                else:
                    assert player_team[uid] == 1
                    x_players[start_frame:end_frame, p, IS_ORANGE] = 1
                x_players[start_frame:end_frame, p, PLAYER_COLS[:-3]] = \
                    player_df[[DEFAULT_FEATURES_STR[c] for c in PLAYER_COLS[:-3]]]

                for frame, pos in player_df[player_df["boost_pickup"] > 0][["pos_x", "pos_y", "pos_z"]].iterrows():
                    boost_id = np.linalg.norm(boost_locations - pos.values, axis=-1).argmin()
                    if boost_locations[boost_id][2] > 72:  # Big boost
                        x_boost[frame: frame + 30 * 10, boost_id, IS_DEMOED] = 1
                    else:  # Small boost
                        x_boost[frame: frame + 30 * 4, boost_id, IS_DEMOED] = 1

        for demo in self.metadata["demos"]:
            frame = demo["frame_number"]
            attacker = demo["attacker_unique_id"]
            victim = demo["victim_unique_id"]

            victim_index = list(self.players).index(victim)

            x_players[frame: frame + 30 * 3, victim_index, IS_DEMOED] = 1

        normalize(x)

        x_s, y_s = [v.copy() for v in x], [v.copy() for v in y]
        swap_teams(x_s, y_s)

        x_m, y_m = [v.copy() for v in x], [v.copy() for v in y]
        swap_left_right(x_m, y_m)

        x_sm, y_sm = [v.copy() for v in x_s], [v.copy() for v in y_s]
        swap_left_right(x_sm, y_sm)

        all_x = []
        all_y = []
        for x_variant, y_variant in ((x, y), (x_s, y_s), (x_m, y_m), (x_sm, y_sm)):
            cls = np.zeros((x_variant[0].shape[0], 1, x_variant[0].shape[2]))
            cls[:, 0, CLS] = 1
            x_variant = np.concatenate(
                [cls] + x_variant,
                axis=1
            )
            y_score = y_variant[0]

            x_variant = x_variant[np.r_[tuple(valid_frames)]]
            y_score = y_score[np.r_[tuple(valid_frames)]]

            np.nan_to_num(x_variant, copy=False)

            all_x.append(x_variant)
            all_y.append(y_score)

        all_x = np.concatenate(all_x, axis=0)
        all_y = np.concatenate(all_y, axis=0)

        return all_x, all_y


class ReplayDataset(IterableDataset):
    def __init__(self, folder):
        self.folder = folder

    def __iter__(self) -> Iterator[T_co]:
        for replay_folder in os.listdir(self.folder):
            path = os.path.join(self.folder, replay_folder)
            if os.path.isdir(path):
                ca = CarballAnalysis(path)
                x, y = ca.get_earl_compatible(n_players=6)
                yield x, y

    def __getitem__(self, index) -> T_co:
        raise NotImplementedError


def main():
    group = "rlcs-x-whqc4pi1kw"
    replay_folder = r"D:\rokutleg\replays\rlcsx"
    output_folder = r"D:\rokutleg\processed\rlcsx"
    os.makedirs(replay_folder, exist_ok=True)

    if len(sys.argv) > 1:
        key = sys.argv[1]
        bc_api = ballchasing.Api(key)
        folder = os.path.join(replay_folder, group)
        os.makedirs(folder, exist_ok=True)

        with open(os.path.join(replay_folder, "replay_info.ijson"), "w") as info:
            for replay in bc_api.get_group_replays(group):
                try:
                    bc_api.download_replay(replay["id"], folder)
                    info.write(json.dumps(replay) + "\n")
                except Exception as e:
                    print(e)

    all_replays = [
        os.path.join(dp, f)
        for dp, dn, fn in os.walk(replay_folder)
        for f in fn
        if f.endswith(".replay")
    ]

    process_map(process_replay, all_replays, [output_folder] * len(all_replays), chunksize=5)

    replay_folder = r"D:\rokutleg\replays\electrum"
    output_folder = r"D:\rokutleg\processed\electrum"
    all_replays = [os.path.join(dp, f) for dp, dn, fn in os.walk(replay_folder) for f in fn]

    process_map(process_replay, all_replays, [output_folder] * len(all_replays), chunksize=5)


def earlify(path):
    if os.path.isdir(path):
        try:
            ca = CarballAnalysis(path)
            x, y = ca.get_earl_compatible(n_players=6)
            return x, y
        except Exception as e:
            print(e)
            pass
        # except FileNotFoundError:
        #     print("Not found error when trying to open:", path)
        # except UnicodeDecodeError:
        #     print("Unicode decode error when trying to open:", path)


def data_gen():
    folder = r"D:\rokutleg\processed\rlcsx"
    files = os.listdir(folder)
    # random.shuffle(files)
    files = [os.path.join(folder, replay_folder) for replay_folder in files]

    # with ProcessPoolExecutor(1) as ex:
    for res in map(earlify, files):
        if res is not None:
            yield res


def train():
    model = NGPModel(EARL(dropout=0.1)).cuda()

    bs = 256
    optimizer = Adam(model.parameters(), lr=3e-4)
    loss_fn = CrossEntropyLoss()
    for epoch in range(1, 100):
        n = 0
        xs = []
        ys = []
        torch.save(model, f"EARL-NGP-{epoch}")
        tot_loss = 0
        b = 0
        t = 0
        for x, y in data_gen():
            xs.append(x)
            ys.append(y)

            if n % 8 == 7:
                xs = torch.from_numpy(np.concatenate(xs)).float().cuda()
                ys = torch.from_numpy(np.concatenate(ys)).long().cuda()

                indices = torch.randperm(xs.shape[0])
                xs = xs[indices]
                ys = ys[indices]

                testing = t % 8 == 0
                # tot_loss = 0
                # b = 0
                for batch in range(0, xs.shape[0] - bs, bs):
                    x_batch = xs[batch: batch + bs]
                    y_batch = ys[batch: batch + bs]

                    if not testing:
                        model.train()
                        optimizer.zero_grad()
                        pred = model(x_batch)
                        loss = loss_fn(pred, y_batch)
                        loss.backward()
                        optimizer.step()
                    else:
                        model.eval()
                        with torch.no_grad():
                            pred = model(x_batch)
                            loss = loss_fn(pred, y_batch)
                        tot_loss += loss.item()
                        b += 1

                if testing:
                    print(tot_loss / (b or 1))
                xs = []
                ys = []
                t += 1

            n += 1
        print(epoch, tot_loss / (b or 1))


if __name__ == '__main__':
    # train()
    main()
    exit(0)
    t0 = time.time()
    ca = CarballAnalysis(r"D:\rokutleg\processed\rlcsx\00b8df8c-0088-4560-91b0-a9b9586cc183")
    for frame in ca.to_rlgym():
        pass
    t1 = time.time()
    print("Hei", t1 - t0)



import numpy as np
from rlgym.utils import RewardFunction
from rlgym.utils.gamestates import PlayerData, GameState


class DiffReward(RewardFunction):
    """
    Calculates difference in reward between time steps
    For instance, it can be used to reward acceleration by measuring change in velocity.
    """
    def __init__(self, reward_function: RewardFunction, negative_slope=1.):
        super().__init__()
        self.reward_function = reward_function
        self.last_values = {}
        self.negative_slope = negative_slope  # Can weight negative values differently

    def reset(self, initial_state: GameState):
        self.last_values = {}

    def _calculate_diff(self, player, rew):
        last = self.last_values.get(player.car_id)
        self.last_values[player.car_id] = rew
        if last is not None:
            ret = rew - last
            return self.negative_slope * ret if ret < 0 else ret
        else:
            return 0

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        rew = self.reward_function.get_reward(player, state, previous_action)
        return self._calculate_diff(player, rew)

    def get_final_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        rew = self.reward_function.get_final_reward(player, state, previous_action)
        return self._calculate_diff(player, rew)



import numpy as np
from rlgym.envs import Match
from rlgym.utils.action_parsers import DiscreteAction
from rlgym.utils.obs_builders import AdvancedObs
from rlgym.utils.reward_functions import DefaultReward
from rlgym.utils.reward_functions.common_rewards import VelocityPlayerToBallReward
from rlgym.utils.state_setters import DefaultState
from rlgym.utils.terminal_conditions.common_conditions import TimeoutCondition, GoalScoredCondition

from rlgym_tools.extra_rewards.multi_model_rewards import MultiModelReward
from rlgym_tools.sb3_utils import SB3MultipleInstanceEnv
from stable_baselines3 import PPO
from stable_baselines3.common.callbacks import CheckpointCallback

from rlgym_tools.sb3_utils.sb3_multi_agent_tools import multi_learn

if __name__ == '__main__':
    frame_skip = 8
    half_life_seconds = 5

    fps = 120 / frame_skip
    gamma = np.exp(np.log(0.5) / (fps * half_life_seconds))  # Quick mafs
    print(f"fps={fps}, gamma={gamma}")

    # models can only be created/loaded after the env, but the model_map and reward_funcs are used to create the env,
    # so we create all of the lists except for the model list here

    # map of players to model indexes, should be of length = n_envs * players_per_env
    model_map = [0, 0, 1, 2, 3, 3, 2, 0]
    # learning mask is the same size as the models list. True for the model to learn.
    learning_mask = [True, False, True, True]
    # some simple rewards for example purposes. reward_funcs should be in the same order as the list of models.
    reward_funcs = [VelocityPlayerToBallReward(), DefaultReward(), VelocityPlayerToBallReward(), DefaultReward()]


    def get_match():
        return Match(
            team_size=2,  # 2v2 for this example because why not
            tick_skip=frame_skip,
            # use the MultiModelReward to handle the distribution of rewards to each model.
            reward_function=MultiModelReward(model_map, reward_funcs),
            self_play=True,
            terminal_conditions=[TimeoutCondition(round(fps * 15)), GoalScoredCondition()],  # Some basic terminals
            obs_builder=AdvancedObs(),  # Not that advanced, good default
            state_setter=DefaultState(),
            action_parser=DiscreteAction()
        )


    env = SB3MultipleInstanceEnv(get_match, 2)  # Start 2 instances

    # Hyperparameters presumably better than default; inspired by original PPO paper
    models = [PPO('MlpPolicy', env) for _ in range(4)]
    for _ in range(4):
        model = PPO('MlpPolicy', env)
        models.append(model)

    # This saves to specified folder with a specified name
    # callbacks is a list the same length as the list of models, in the same order.
    callbacks = [CheckpointCallback(round(1_000_000 / env.num_envs), save_path="policy", name_prefix=f"multi_{n}") for n
                 in range(4)]

    # It can be a good idea to call multi-learn multiple times in a loop, modifying model_map in-place (if it is not
    # done in-place, the reward functions will desync) to get extra speed by limiting the number of models training at
    # once. (more separate models training = more calculation time each step)
    multi_learn(
        models=models,  # the list of models that will be used
        total_timesteps=10_000_000,  # total timestamps that will be trained for
        env=env,
        callbacks=callbacks,  # list of callbacks, one for each model in the list of models
        num_players=8,  # team_size * num_instances * 2
        model_map=model_map,  # mapping of models to players.
        learning_mask=learning_mask
    )



import logging
import os

import numpy as np
import ray
from gym.spaces import Box
from ray.rllib.agents.ppo import PPOTorchPolicy, PPOTrainer
from ray.tune import register_env, tune

import rlgym
from rlgym.envs import Match
from rlgym.gym import Gym
from rlgym.utils.action_parsers import DiscreteAction
from rlgym.utils.obs_builders import AdvancedObs
from rlgym.utils.reward_functions.common_rewards import VelocityReward
from rlgym.utils.state_setters import DefaultState
from rlgym.utils.terminal_conditions.common_conditions import TimeoutCondition, GoalScoredCondition
from rlgym_tools.rllib_utils import RLLibEnv

if __name__ == '__main__':
    # ray.init(namespace="rlgym", logging_level=logging.DEBUG)
    ray.init(address='127.0.0.1:6379', _redis_password='5241590000000000', logging_level=logging.DEBUG)
    # ray.init(address='ray://127.0.0.1:6379')
    # tune.run(PPOTrainer, config={"env": "CartPole-v0", "num_workers": 2, "num_cpus_per_worker": 1,
    #                              "num_envs_per_worker": 1})  # "log_level": "INFO" for verbose,
    # "framework": "tfe"/"tf2" for eager,
    # "framework": "torch" for PyTorch
    # exit(0)
    # ray.init(logging_level=logging.DEBUG, memory=4e9, object_store_memory=4e9)

    # ray start --head --port=6379 --ray-client-server-port=6968 --gcs-server-port=6969 --dashboard-port=8265
    # docker run --shm-size=4.07gb -i -p 6379:6379 -p 8265:8265 -p 6969:6969 -p 6968:6968 b926c8b4c92b

    rl_path = r"C:\Program Files\Epic Games\rocketleague\Binaries\Win64\RocketLeague.exe"


    def create_env():
        return RLLibEnv(
            rlgym.make(
                self_play=True,
                obs_builder=AdvancedObs(),
                reward_fn=VelocityReward(negative=True),
                state_setter=DefaultState(),
                action_parser=DiscreteAction(),
                terminal_conditions=[TimeoutCondition(15 * 120 // 8), GoalScoredCondition()]
            )
        )


    register_env("RLGym", create_env)

    policy = PPOTorchPolicy, Box(-np.inf, np.inf, (107,)), Box(-1.0, 1.0, (8,)), {}
    # policy = PPOTorchPolicy, Box(-np.inf, np.inf, (4,)), Discrete(2), {}

    tune.run(
        PPOTrainer,
        config={
            "env": "RLGym",  # "CartPole-v0",
            "multiagent": {
                "policies": {"ppo_policy": policy},
                "policy_mapping_fn": (lambda agent_id, **kwargs: "ppo_policy"),
                "policies_to_train": ["ppo_policy"],
            },
            # "env_config": {
            #     "num_agents": 2
            # },
            "model": {
                "vf_share_layers": True,
            },
            # Use GPUs iff `RLLIB_NUM_GPUS` env var set to > 0.
            # "num_gpus": int(os.environ.get("RLLIB_NUM_GPUS", "0")),
            "num_workers": 1,
            "num_cpus_per_worker": 1,
            "num_envs_per_worker": 1,
            "remote_worker_envs": True,
            "sample_async": True,
            "framework": "torch"
        },
    )

    # for i in range(1000000):
    #     print("== Iteration", i, "==")
    #
    #     # improve the PPO policy
    #     print("-- PPO --")
    #     result_ppo = ppo_trainer.train()
    #     print(pretty_print(result_ppo))
    # print("Done training")

    # # This does not work for some reason
    # tune.run(
    #     "PPO",
    #     config={
    #         "env": MultiAgentCartPole,
    #         "env_config": {
    #             "num_agents": 2
    #         },
    #         "multiagent": {
    #             "policies": {"ppo_policy": policy},
    #             "policy_mapping_fn": (lambda agent_id, **kwargs: "ppo_policy"),
    #             "policies_to_train": ["ppo_policy"],
    #         },
    #         "model": {
    #             "vf_share_layers": True,
    #         },
    #         # Use GPUs iff `RLLIB_NUM_GPUS` env var set to > 0.
    #         "num_gpus": int(os.environ.get("RLLIB_NUM_GPUS", "0")),
    #         "num_workers": 2,
    #         "num_cpus_per_worker": 1,
    #         "num_workers_per_env": 2,
    #         "remote_worker_envs": True,
    #         "sample_async": True,
    #         "framework": "torch"
    #     }
    # )



import numpy as np

from rlgym.utils import RewardFunction, math
from rlgym.utils.common_values import BLUE_TEAM, ORANGE_TEAM, ORANGE_GOAL_BACK, \
    BLUE_GOAL_BACK, BALL_MAX_SPEED, BACK_WALL_Y, BALL_RADIUS, BACK_NET_Y
from rlgym.utils.gamestates import GameState, PlayerData


class LiuDistanceBallToGoalReward(RewardFunction):
    def __init__(self, own_goal=False):
        super().__init__()
        self.own_goal = own_goal

    def reset(self, initial_state: GameState):
        pass

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        if player.team_num == BLUE_TEAM and not self.own_goal \
                or player.team_num == ORANGE_TEAM and self.own_goal:
            objective = np.array(ORANGE_GOAL_BACK)
        else:
            objective = np.array(BLUE_GOAL_BACK)

        # Compensate for moving objective to back of net
        dist = np.linalg.norm(state.ball.position - objective) - (BACK_NET_Y - BACK_WALL_Y + BALL_RADIUS)
        return np.exp(-0.5 * dist / BALL_MAX_SPEED)  # Inspired by https://arxiv.org/abs/2105.12196


class VelocityBallToGoalReward(RewardFunction):
    def __init__(self, own_goal=False, use_scalar_projection=False):
        super().__init__()
        self.own_goal = own_goal
        self.use_scalar_projection = use_scalar_projection

    def reset(self, initial_state: GameState):
        pass

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        if player.team_num == BLUE_TEAM and not self.own_goal \
                or player.team_num == ORANGE_TEAM and self.own_goal:
            objective = np.array(ORANGE_GOAL_BACK)
        else:
            objective = np.array(BLUE_GOAL_BACK)

        vel = state.ball.linear_velocity
        pos_diff = objective - state.ball.position
        if self.use_scalar_projection:
            # Vector version of v=d/t <=> t=d/v <=> 1/t=v/d
            # Max value should be max_speed / ball_radius = 2300 / 94 = 24.5
            # Used to guide the agent towards the ball
            inv_t = math.scalar_projection(vel, pos_diff)
            return inv_t
        else:
            # Regular component velocity
            norm_pos_diff = pos_diff / np.linalg.norm(pos_diff)
            vel /= BALL_MAX_SPEED
            return float(np.dot(norm_pos_diff, vel))


class BallYCoordinateReward(RewardFunction):
    def __init__(self, exponent=1):
        # Exponent should be odd so that negative y -> negative reward
        self.exponent = exponent

    def reset(self, initial_state: GameState):
        pass

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        if player.team_num == BLUE_TEAM:
            return (state.ball.position[1] / (BACK_WALL_Y + BALL_RADIUS)) ** self.exponent
        else:
            return (state.inverted_ball.position[1] / (BACK_WALL_Y + BALL_RADIUS)) ** self.exponent



"""
    The Rocket League gym environment.
"""

from time import sleep
from typing import List, Union, Tuple, Dict, Any

import numpy as np
from gym import Env

from rlgym.communication import CommunicationHandler, Message
from rlgym.communication.communication_exception_handler import CrashPolicy
from rlgym.gamelaunch import launch_rocket_league, run_injector, page_rocket_league, LaunchPreference


class Gym(Env):
    def __init__(self, match, pipe_id=0, launch_preference=LaunchPreference.EPIC, use_injector=False,
                 force_paging=False, crash_policy=CrashPolicy.ATTEMPT_RECOVER):
        super().__init__()

        self._match = match
        self.observation_space = match.observation_space
        self.action_space = match.action_space

        self._launch_preference = launch_preference
        self._use_injector = use_injector
        self._force_paging = force_paging

        self._crash_policy = crash_policy

        self._comm_handler = CommunicationHandler()
        self._local_pipe_name = CommunicationHandler.format_pipe_id(pipe_id)
        self._local_pipe_id = pipe_id

        self._game_process = None

        self._open_game()
        self._setup_plugin_connection()

        if self._force_paging:
            self._page_client()

        self._prev_state = None

    def _open_game(self):
        print("Launching Rocket League, make sure bakkesmod is running.")
        # Game process is only set if epic version is used or launched with path_to_rl
        self._game_process = launch_rocket_league(self._local_pipe_name, self._launch_preference)

        if self._use_injector:
            sleep(3)
            run_injector()

    def _setup_plugin_connection(self):
        self._comm_handler.open_pipe(self._local_pipe_name)
        self._comm_handler.send_message(header=Message.RLGYM_CONFIG_MESSAGE_HEADER, body=self._match.get_config())
        self._comm_handler.send_message(header=Message.RLGYM_RESET_GAME_STATE_MESSAGE_HEADER,
                                        body=self._match.get_reset_state())

    def _page_client(self) -> bool:
        if self._game_process is None:
            print("Forced paging is only supported for the epic games version")
            return False
        else:
            print("Forcing Rocket League to page unused memory. PID:", self._game_process.pid)
            return page_rocket_league(rl_pid=self._game_process.pid)

    def reset(self, return_info=False) -> Union[List, Tuple]:
        """
        The environment reset function. When called, this will reset the state of the environment and objects in the game.
        This should be called once when the environment is initialized, then every time the `done` flag from the `step()`
        function is `True`.
        """

        state_str = self._match.get_reset_state()

        exception = self._comm_handler.send_message(header=Message.RLGYM_RESET_GAME_STATE_MESSAGE_HEADER,
                                                    body=state_str)
        if exception is not None:
            self._attempt_recovery()
            exception = self._comm_handler.send_message(header=Message.RLGYM_RESET_GAME_STATE_MESSAGE_HEADER,
                                                        body=state_str)
            if exception is not None:
                import sys
                print("!UNABLE TO RECOVER ROCKET LEAGUE!\nEXITING")
                sys.exit(-1)

        state = self._receive_state()
        self._match.episode_reset(state)
        self._prev_state = state

        obs = self._match.build_observations(state)
        if return_info:
            info = {
                'state': state,
                'result': self._match.get_result(state)
            }
            return obs, info
        return obs

    def step(self, actions: Any) -> Tuple[List, List, bool, Dict]:
        """
        The step function will send the list of provided actions to the game, then advance the game forward by `tick_skip`
        physics ticks using that action. The game is then paused, and the current state is sent back to RLGym. This is
        decoded into a `GameState` object, which gets passed to the configuration objects to determine the rewards,
        next observation, and done signal.

        :param actions: An object containing actions, in the format specified by the `ActionParser`.
        :return: A tuple containing (obs, rewards, done, info)
        """

        actions = self._match.parse_actions(actions, self._prev_state)
        actions_sent = self._send_actions(actions)

        received_state = self._receive_state()

        # If, for any reason, the state is not successfully received, we do not want to just crash the API.
        # This will simply pretend that the state did not change and advance as though nothing went wrong.
        if received_state is None:
            print("FAILED TO RECEIEVE STATE! FALLING TO", self._prev_state)
            state = self._prev_state
        else:
            state = received_state

        obs = self._match.build_observations(state)
        done = self._match.is_done(state) or received_state is None or not actions_sent
        reward = self._match.get_rewards(state, done)
        self._prev_state = state

        info = {
            'state': state,
            'result': self._match.get_result(state)
        }

        return obs, reward, done, info

    def close(self):
        """
        Disconnect communication with the Bakkesmod plugin and close the game. This should only be called if you are finished
        with your current RLGym environment instance.
        """
        self._comm_handler.close_pipe()
        if self._game_process is not None:
            self._game_process.terminate()

    def _receive_state(self):
        # print("Waiting for state...")
        message, exception = self._comm_handler.receive_message(header=Message.RLGYM_STATE_MESSAGE_HEADER)
        if exception is not None:
            self._attempt_recovery()
            return None

        if message is None:
            return None
        # print("GOT MESSAGE\n HEADER: '{}'\nBODY: '{}'\n".format(message.header, message.body))
        if message.body is None:
            return None

        return self._match.parse_state(message.body)

    def _send_actions(self, actions):
        assert isinstance(actions, np.ndarray), "Invalid action type, action must be of type np.ndarray(n, 8)."
        assert len(actions.shape) == 2, "Invalid action shape, shape must be of the form (n, 8)."
        assert actions.shape[-1] == 8, "Invalid action shape, last dimension must be 8."

        actions_formatted = self._match.format_actions(actions)
        exception = self._comm_handler.send_message(header=Message.RLGYM_AGENT_ACTION_IMMEDIATE_RESPONSE_MESSAGE_HEADER,
                                                    body=actions_formatted)
        if exception is not None:
            self._attempt_recovery()
            return False
        return True

    def _attempt_recovery(self):
        if self._crash_policy == CrashPolicy.RAISE:
            raise EnvironmentError
        print("!ROCKET LEAGUE HAS CRASHED!\nATTEMPTING RECOVERY")
        import os
        import time
        self.close()
        proc_list = os.popen('wmic process get description, processid').read()
        num_instances = proc_list.count("RocketLeague.exe")
        wait_time = 2 * num_instances

        print("Discovered {} existing Rocket League processes. Waiting {} seconds before attempting to open "
              "a new one.".format(num_instances, wait_time))

        time.sleep(wait_time)
        self._open_game()
        self._setup_plugin_connection()
        if self._force_paging:
            self._page_client()
        if self._crash_policy == CrashPolicy.ATTEMPT_RECOVER_THEN_RAISE:
            raise EnvironmentError



import math
import numpy as np
import gym.spaces
from rlgym.utils import common_values
from rlgym.utils.gamestates import GameState
from rlgym.utils.action_parsers import ActionParser


class ContinuousAction(ActionParser):
    """
        Simple continuous action space. Operates in the range -1 to 1, even for the binary actions which are converted back to binary later.
        This is for improved compatibility, stable baselines doesn't support tuple spaces right now.
    """

    def __init__(self):
        super().__init__()

    def get_action_space(self) -> gym.spaces.Space:
        # return gym.spaces.Tuple((gym.spaces.Box(-1, 1, shape=(5,)), gym.spaces.MultiBinary(3)))
        return gym.spaces.Box(-1, 1, shape=(common_values.NUM_ACTIONS,))

    def parse_actions(self, actions: np.ndarray, state: GameState) -> np.ndarray:
        actions = actions.reshape((-1, 8))

        actions[..., :5] = actions[..., :5].clip(-1, 1)
        # The final 3 actions handle are jump, boost and handbrake. They are inherently discrete so we convert them to either 0 or 1.
        actions[..., 5:] = actions[..., 5:] > 0

        return actions



import logging

import carball as cb
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from carball.controls.controls import ControlsCreator

from ..dataset.create_dataset import replay_to_dfs, convert_dfs, normalize


def sigmoid(x):
    return 1 / (1 + np.exp(-x))


def inv_sigmoid(x):
    return np.log(x / (1 - x))


def predict_boost_values(rp, model):
    if isinstance(rp, str):
        rp = cb.analyze_replay_file(rp,
                                    controls=ControlsCreator(),
                                    logging_level=logging.CRITICAL)

    pp = replay_to_dfs(rp, skip_ties=False)
    cpx, cpy = convert_dfs(pp, frame_mode=1)
    normalize(cpx)

    results = np.zeros((cpx[0].shape[0], len(pp["players"]), 101))

    # xs = np.zeros((cpx[0].shape[0], len(pp["players"])))
    # ys = np.zeros((cpx[0].shape[0], len(pp["players"])))

    model.eval()
    with torch.no_grad():
        for p, row in pp["players"].iterrows():
            print(p)
            cpxc = [np.copy(v) for v in cpx]

            # # For making histogram
            # preds = model(*[torch.from_numpy(v).float().cuda() for v in cpxc])
            # preds = preds[0].cpu().detach().numpy()
            # preds = preds[:, 1] - preds[:, 0]
            # if row["color"] == "blue":
            #     preds = -preds
            # xs[:, p] = cpxc[2][:, p, 19]
            # ys[:, p] = preds

            # For changing the boost amount at every step
            for b in range(101):
                print("\t", b)
                cpxc[2][:, p, 19] = b / 100
                cpxc[2][:, p, 19][cpxc[2][:, p, 20] == 1] = 0
                preds = model(*[torch.from_numpy(v).float().cuda() for v in cpxc])
                preds = preds[0].cpu().detach().numpy()
                preds = preds[:, 1] - preds[:, 0]
                if row["color"] == "blue":
                    preds = -preds
                results[:, p, b] = preds

    return results


def make_summary(rp, model):
    if isinstance(rp, str):
        rp = cb.analyze_replay_file(rp,
                                    logging_level=logging.CRITICAL)

    pp = replay_to_dfs(rp, skip_ties=False)

    cpx, cpy = convert_dfs(pp, frame_mode=1)
    # print(cpy)

    normalize(cpx)

    pred_df = pd.DataFrame(index=pp["frames"].index)

    model.eval()

    with torch.no_grad():

        def get_predictions(x_data):
            preds = None
            bs = 1024
            lo, hi = 0, bs
            for batch in range(bs, len(x_data[0]) + bs, bs):
                hi = min(batch, len(x_data[0]))

                batch_x = [v[lo:hi] for v in x_data]
                batch_preds_b = model(*[torch.from_numpy(v).float().cuda() for v in batch_x])
                batch_preds_b = [bp.cpu().detach().numpy() for bp in batch_preds_b]

                # swap_teams(batch_x)
                # batch_preds_s = model(*[torch.from_numpy(v).float().cuda() for v in batch_x])
                # batch_preds_s = [bp.cpu().detach().numpy() for bp in batch_preds_s]
                #
                # swap_left_right(batch_x)
                # batch_preds_sm = model(*[torch.from_numpy(v).float().cuda() for v in batch_x])
                # batch_preds_sm = [bp.cpu().detach().numpy() for bp in batch_preds_sm]
                #
                # swap_teams(batch_x)
                # batch_preds_m = model(*[torch.from_numpy(v).float().cuda() for v in batch_x])
                # batch_preds_m = [bp.cpu().detach().numpy() for bp in batch_preds_m]
                #
                # batch_preds = [(b + s + m + sm) / 4 for b, s, m, sm in
                #                zip(batch_preds_b, batch_preds_s, batch_preds_m, batch_preds_sm)]
                batch_preds = batch_preds_b

                if preds is None:
                    preds = batch_preds
                else:
                    preds = [np.concatenate([p, bp]) for p, bp in zip(preds, batch_preds)]
                lo = hi
            return preds

        # 1. Get predictions with all players
        preds = get_predictions(cpx)
        pred_df["pred"] = preds[0][:, 0] - preds[0][:, 1]

        for i, row in pp["players"].iterrows():
            color = row["color"]
            sgn = 1 if color == "blue" else -1
            cpxc = [np.copy(v) for v in cpx]
            cpxc[2] = cpxc[2][:, [k for k in range(cpx[2].shape[1]) if k != i]]

            # 2. Get predictions with one player removed
            preds = get_predictions(cpxc)
            pred_df[f"{i}/pred_removed"] = preds[0][:, 0] - preds[0][:, 1]

            cpxc = [np.copy(v) for v in cpx]
            cpxc[2] = cpxc[2][:, [i]]

            # 3. Get predictions with teammates removed
            preds = get_predictions(cpxc)
            pred_df[f"{i}/pred_solo"] = preds[0][:, 0] - preds[0][:, 1]

            # pred_df[f"{identifier}/pred"] = sigmoid(sgn * pred_df["pred"]) * \
            #                                 sigmoid(sgn * pred_df[f"{identifier}/pred_solo"]) / \
            #                                 sigmoid(sgn * pred_df[f"{identifier}/pred_removed"])

            pred_df[f"{i}/pred"] = sgn * (
                    pred_df["pred"] - pred_df[f"{i}/pred_removed"])

            # pred_df[f"{identifier}/pred"] = sgn * (
            #         pred_df["pred"] - pred_df[f"{identifier}/pred_removed"] + pred_df[f"{identifier}/pred_solo"])

        return pred_df, pp


def plot_replay(rp_path, model, plot_players=True, invert=None, plot_goal_distance=False):
    if plot_players is True:
        plot_players = ("blue", "orange")
    elif plot_players is False:
        plot_players = ()
    elif plot_players == "blue":
        plot_players = ("blue",)
    elif plot_players == "orange":
        plot_players = ("orange",)
    else:
        plot_players = ()

    rp = cb.analyze_replay_file(rp_path,
                                controls=ControlsCreator(),
                                logging_level=logging.CRITICAL)

    id_team = {p.online_id: ["blue", "orange"][p.team.is_orange] for p in rp.game.players}
    id_name = {p.online_id: p.name for p in rp.game.players}

    pred_df, dfs = make_summary(rp, model)

    if invert is None:
        d = 0
        for goal in rp.game.goals:
            if goal.player.is_orange:
                d += 1
            else:
                d -= 1
        invert = d < 0

    # pp["pred"][pp["ball/ball/vel_y"] == 0] = float("nan")
    # pp["pred"][pp["ball/ball/vel_y"].isna()] = float("nan")

    fig, ax = plt.subplots(figsize=((6.4 * dfs["frames"].index[-1] + 1000) / 5000, 4.8), dpi=400,
                           constrained_layout=True)

    ax.hlines(0.5, xmin=0, xmax=dfs["frames"].index[-1], color="black", linestyle="dashed")

    blue_shots = []
    orange_shots = []

    blue_saves = []
    orange_saves = []

    for hit in rp.protobuf_game.game_stats.hits:
        if hit.shot:
            if id_team[hit.player_id.id] == "blue":
                blue_shots.append(hit.frame_number)
            else:
                orange_shots.append(hit.frame_number)
        if hit.save:
            if id_team[hit.player_id.id] == "blue":
                blue_saves.append(hit.frame_number)
            else:
                orange_saves.append(hit.frame_number)

    blue_ymin, blue_ymax, orange_ymin, orange_ymax = (0.5, 1, 0, 0.5) if invert else (0, 0.5, 0.5, 1)

    ax.vlines(blue_shots, ymin=blue_ymin, ymax=blue_ymax,
              color="m", linestyle="dotted", label="Shot")
    ax.vlines(orange_shots, ymin=orange_ymin, ymax=orange_ymax,
              color="m", linestyle="dotted")

    ax.vlines(blue_saves, ymin=orange_ymin, ymax=orange_ymax,
              color="c", linestyle="dotted", label="Save")
    ax.vlines(orange_saves, ymin=blue_ymin, ymax=blue_ymax,
              color="c", linestyle="dotted")

    if plot_goal_distance:
        goal_location = np.array([0, -5120 if invert else 5120, 642 / 2])
        distances = np.linalg.norm(
            np.tile(goal_location, len(dfs["frames"])).reshape((len(dfs["frames"]), 3)) - dfs["frames"].filter(
                regex="ball/pos_").values, axis=-1)
        ax.plot(dfs["frames"].index, distances / 11216, label="Ball-Goal Distance")

    pred = sigmoid(pred_df["pred"])
    ax.plot(dfs["frames"].index, pred if invert else 1 - pred, color="blue", label="Advantage")

    blue_colors = ["springgreen", "mediumseagreen", "green"]
    orange_colors = ["lightcoral", "indianred", "brown"]
    # plt.show()

    b = o = 0
    for i, row in dfs["players"].iterrows():
        name = row["name"]
        color = row["color"]

        avg_score = np.exp(pred_df[f"{i}/pred"].mean())
        print(name, avg_score)

        # pred_df[f"{identifier}/pred"].plot.hist(title=name, bins=100)
        # plt.show()

        pred_col = sigmoid(pred_df[f"{i}/pred"])
        if color == "blue" and "blue" in plot_players:
            ax.plot(dfs["frames"].index, pred_col if invert else 1 - pred_col, color=blue_colors[b],
                    label=name, linewidth=0.5)
            b += 1
        elif color == "orange" and "orange" in plot_players:
            ax.plot(dfs["frames"].index, 1 - pred_col if invert else pred_col, color=orange_colors[o], label=name,
                    linewidth=0.5)
            o += 1

    rallies = dfs["rallies"]
    ax.vlines(rallies[rallies["team"].notna()]["end_frame"], ymin=0, ymax=1, color="red", label="Goal")
    ax.set_xlim(0, dfs["frames"].index[-1])
    ax.set_ylim(0, 1)
    ax.set_xlabel("Frame")
    ax.set_ylabel("Orange <-> Blue" if invert else "Blue <-> Orange")
    ax.set_title(rp.game.name)
    ax.legend(bbox_to_anchor=(1.01, 1), loc='upper left')

    secax = ax.twiny()
    new_tick_locations = []
    new_ticks = []
    taken = None
    for frame in reversed(rp.game.frames.index):
        left = rp.game.frames["seconds_remaining"].loc[frame]
        if np.isnan(left):
            continue
        left = int(left)
        if left % 10 == 0 and left != taken:
            new_tick_locations.append(frame)
            new_ticks.append("{0:.0f}:{1:02.0f}".format(*divmod(rp.game.frames["seconds_remaining"].loc[frame], 60)))
            taken = left
    secax.set_xlim(*ax.get_xlim())
    secax.set_xticks(new_tick_locations)
    secax.set_xticklabels(new_ticks)
    secax.set_xlabel("Time")

    ax.grid()
    # df["pred"].plot(x=df["game/info/frame"], color="blue")
    return fig, dfs


if __name__ == '__main__':
    import torch

    mdl = torch.load("../../out/models/EARLReplayModel(EARL(n_dims=256,n_layers=8,n_heads=8))_trained.model.ep3").cuda()
    # mdl = EARLReplayModel().cuda()
    nrg_worlds = r"C:\Users\rolv_\Downloads\012194B14489DD6AD776F0A877C53C05.replay"
    bds = r"C:\Users\rolv_\Downloads\ffaed1b9-ac42-4923-9e5b-66eaeef135c1.replay"
    # eltaco_loss = r"C:\Users\rolv_\Downloads\987fe5bd-4b18-4678-8a4f-afabb443b3f9.replay"
    eltaco_win = r"C:\Users\rolv_\Downloads\62a59623-081f-4caf-8eaa-b9b3b5bbb1fd.replay"

    # xs, ys = predict_boost_values(bds, mdl)
    # df = pd.DataFrame({"x": xs * 100, "y": ys})
    # means = df.groupby(df["x"].round()).mean()
    # means = (means["y"] - means["y"][0.]) / (means["y"][1.] - means["y"][0.])
    # plt.plot(means)
    # plt.show()

    boost_values = predict_boost_values(bds, mdl)
    np.save("boost_values.npy", boost_values)

    # boost_values = np.load("boost_values.npy")
    # boost_values = sigmoid(boost_values)
    boost_values = (boost_values - boost_values[:, :, [0]]) / (boost_values[:, :, [100]] - boost_values[:, :, [0]])
    boost_values = np.nanmean(boost_values, (0, 1), )
    print("{" + ", ".join(f"({b / 100}, {boost_values[b]})" for b in range(101)) + "}")
    plt.plot(boost_values)
    plt.show()

    # f, p = plot_replay(nrg_worlds, mdl)
    # f.show()



from rlgym.utils.state_setters import StateSetter
from rlgym.utils.state_setters import StateWrapper
import random
import numpy as np


class DefaultState(StateSetter):

    SPAWN_BLUE_POS = [[-2048, -2560, 17], [2048, -2560, 17],
                      [-256, -3840, 17], [256, -3840, 17], [0, -4608, 17]]
    SPAWN_BLUE_YAW = [0.25 * np.pi, 0.75 * np.pi,
                      0.5 * np.pi, 0.5 * np.pi, 0.5 * np.pi]
    SPAWN_ORANGE_POS = [[2048, 2560, 17], [-2048, 2560, 17],
                        [256, 3840, 17], [-256, 3840, 17], [0, 4608, 17]]
    SPAWN_ORANGE_YAW = [-0.75 * np.pi, -0.25 *
                        np.pi, -0.5 * np.pi, -0.5 * np.pi, -0.5 * np.pi]

    def __init__(self):
        super().__init__()

    def reset(self, state_wrapper: StateWrapper):
        """
        Modifies state_wrapper values to emulate a randomly selected default kickoff.

        :param state_wrapper: StateWrapper object to be modified with desired state values.
        """
        # possible kickoff indices are shuffled
        spawn_inds = [0, 1, 2, 3, 4]
        random.shuffle(spawn_inds)

        blue_count = 0
        orange_count = 0
        for car in state_wrapper.cars:
            pos = [0,0,0]
            yaw = 0
            # team_num = 0 = blue team
            if car.team_num == 0:
                # select a unique spawn state from pre-determined values
                pos = self.SPAWN_BLUE_POS[spawn_inds[blue_count]]
                yaw = self.SPAWN_BLUE_YAW[spawn_inds[blue_count]]
                blue_count += 1
            # team_num = 1 = orange team
            elif car.team_num == 1:
                # select a unique spawn state from pre-determined values
                pos = self.SPAWN_ORANGE_POS[spawn_inds[orange_count]]
                yaw = self.SPAWN_ORANGE_YAW[spawn_inds[orange_count]]
                orange_count += 1
            # set car state values
            car.set_pos(*pos)
            car.set_rot(yaw=yaw)
            car.boost = 0.33



from rlgym.utils.state_setters.wrappers import PhysicsWrapper
from rlgym.utils.gamestates import PlayerData
import numpy as np



class CarWrapper(PhysicsWrapper):

    def __init__(self, team_num: int = -1, id: int = -1, player_data: PlayerData = None):
        """
        CarWrapper constructor. Under most circumstances, users should not expect to instantiate their own CarWrapper objects.

        :param team_num: Integer indicating 0 for blue and 1 for orange.
        :param id: Integer indicating the spectator ID assigned to the car.
        :param player_data: PlayerData object for values to be copied from.

        NOTE: team_num and id will be ignored if a PlayerData object is passed.
        """
        if player_data is None:
            super().__init__()
            self.position = np.asarray([id * 100, 0, 17])
            self.rotation: np.ndarray = np.zeros(3)
            self.team_num: int = team_num
            self.id: int = id
            self.boost: float = 0
        else:
            super().__init__(phys_obj=player_data.car_data)
            self._read_from_player_data(player_data)

    def _read_from_player_data(self, player_data: PlayerData):
        """
        A function to modify CarWrapper values from values in a PlayerData object.
        """
        self.rotation = player_data.car_data.euler_angles()
        self.team_num = player_data.team_num
        self.id = player_data.car_id
        self.boost = player_data.boost_amount

    def set_rot(self, pitch: float = None, yaw: float = None, roll: float = None):
        """
        Sets rotation in terms of euler angles.
        """
        if pitch is not None:
            self.rotation[0] = pitch
        if yaw is not None:
            self.rotation[1] = yaw
        if roll is not None:
            self.rotation[2] = roll

    def _encode(self) -> list:
        """
        Function called by a StateWrapper to produce a state string.

        :return: String containing value data.
        """
        encoded = np.concatenate(((self.id,), self.position, self.linear_velocity,
                                    self.angular_velocity, self.rotation, (self.boost,)))

        return encoded.tolist()


import json
import os
from typing import Tuple, Optional, List

import numpy as np
from rlgym.utils import RewardFunction
from rlgym.utils.gamestates import PlayerData, GameState
from rlgym.utils.reward_functions import CombinedReward
from stable_baselines3.common.callbacks import BaseCallback
from stable_baselines3.common.logger import Logger


class SB3LogReward(RewardFunction):
    """
    Simple reward function for logging individual rewards to a custom Logger.
    """

    def __init__(self, logger: Logger, reward_function: RewardFunction):
        super().__init__()
        self.logger = logger
        self.reward_function = reward_function
        self.reward_sum = 0
        self.episode_steps = 0
        self.global_steps = 0

    def reset(self, initial_state: GameState):
        if self.episode_steps > 0:
            rew_fn_type = type(self.reward_function)
            mean_reward = self.reward_sum / self.episode_steps
            if rew_fn_type.__str__ is not object.__str__:
                self.logger.record(f"{self.reward_function}/ep_rew_mean", mean_reward)
            else:
                self.logger.record(f"{rew_fn_type.__name__}/ep_rew_mean", mean_reward)
            self.logger.dump(self.global_steps)
            self.reward_sum = 0
            self.episode_steps = 0
            self.global_steps += 1

        self.reward_function.reset(initial_state)

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        rew = self.reward_function.get_reward(player, state, previous_action)
        self.reward_sum += rew
        self.episode_steps += 1
        return rew

    def get_final_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        rew = self.reward_function.get_final_reward(player, state, previous_action)
        self.reward_sum += rew
        self.episode_steps += 1
        return rew


class SB3CombinedLogReward(CombinedReward):

    def __init__(
            self,
            reward_functions: Tuple[RewardFunction, ...],
            reward_weights: Optional[Tuple[float, ...]] = None,
            file_location: str = 'combinedlogfiles'
    ):
        """
        Creates the combined reward using multiple rewards, and a potential set
        of weights for each reward. Will also log the weighted rewards to
        the model's logger if a SB3CombinedLogRewardCallback is provided to the
        learner.

        :param reward_functions: Each individual reward function.
        :param reward_weights: The weights for each reward.
        :param file_location: The path to the directory that will be used to
        transfer reward info
        """
        super().__init__(reward_functions, reward_weights)

        # Make sure there is a folder to dump to
        os.makedirs(file_location, exist_ok=True)
        self.file_location = f'{file_location}/rewards.txt'
        self.lockfile = f'{file_location}/reward_lock'

        # Initiates the array that will store the episode totals
        self.returns = np.zeros(len(self.reward_functions))

        # Obtain the lock
        while True:
            try:
                open(self.lockfile, 'x')
                break
            except FileExistsError:
                pass
            except PermissionError:
                pass
            except Exception as e:
                print(f'Error obtaining lock in SB3CombinedLogReward.__init__:\n{e}')

        # Empty the file by opening in w mode
        with open(self.file_location, 'w') as f:
            pass

        # Release the lock
        try:
            os.remove(self.lockfile)
        except FileNotFoundError:
            print('No lock to release! ')

    def reset(self, initial_state: GameState):
        self.returns = np.zeros(len(self.reward_functions))
        super().reset(initial_state)

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        rewards = [
            func.get_reward(player, state, previous_action)
            for func in self.reward_functions
        ]

        self.returns += [a * b for a, b in zip(rewards, self.reward_weights)]  # store the rewards

        return float(np.dot(self.reward_weights, rewards))

    def get_final_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        rewards = [
            func.get_final_reward(player, state, previous_action)
            for func in self.reward_functions
        ]
        # Add the rewards to the cumulative totals with numpy broadcasting
        self.returns += [a * b for a, b in zip(rewards, self.reward_weights)]

        # Obtain the lock
        while True:
            try:
                open(self.lockfile, 'x')
                break
            except FileExistsError:
                pass
            except PermissionError:
                pass
            except Exception as e:
                print(f'Error obtaining lock in SB3CombinedLogReward.get_final_reward:\n{e}')

        # Write the rewards to file and reset
        with open(self.file_location, 'a') as f:
            f.write('\n' + json.dumps(self.returns.tolist()))

        # reset the episode totals
        self.returns = np.zeros(len(self.reward_functions))

        # Release the lock
        try:
            os.remove(self.lockfile)
        except FileNotFoundError:
            print('No lock to release! ')

        return float(np.dot(self.reward_weights, rewards))


class SB3CombinedLogRewardCallback(BaseCallback):
    def __init__(self, reward_names: List[str], file_location: str = 'combinedlogfiles'):
        """
        Callback to log the data from a SB3CombinedLogReward to the
        same log as the model.

        :param reward_names: List of names that the logger will use for
        each reward.
        :param file_location: The path to the directory that will be used to
        transfer reward info
        """
        super().__init__()
        self.reward_names = reward_names
        self.file_location = file_location + '/rewards.txt'
        self.lockfile = file_location + '/reward_lock'

    def _on_step(self) -> bool:
        return True

    def _on_rollout_end(self) -> None:
        returns = []

        # Obtain the lock
        while True:
            try:
                open(self.lockfile, 'x')
                break
            except FileExistsError:
                pass
            except PermissionError:
                pass
            except Exception as e:
                print(f'Error obtaining lock in SB3CombinedLogRewardCallback._on_rollout_end:\n{e}')

        # Read the file into returns
        with open(self.file_location, 'r') as f:
            for line in f:
                line = line.strip()
                if line:
                    try:
                        line = json.loads(line)
                        returns.append(line)

                    except Exception as e:
                        print(f'Exception loading line {line}:\n\t{e}')

        # Release the lock
        try:
            os.remove(self.lockfile)
        except FileNotFoundError:
            print('No lock to release! ')

        # Make returns into a numpy array so we can make use of numpy features
        returns = np.array(returns)

        # Log each reward
        for n in range(returns.shape[1]):
            try:
                name = self.reward_names[n]
            except IndexError:
                name = f'reward_{n}'
            self.model.logger.record_mean('rewards/' + name, np.mean(returns[:, n]))



from rlgym.utils import TerminalCondition
from rlgym.utils.gamestates import GameState


class GameCondition(TerminalCondition):  # Mimics a Rocket League game
    def __init__(self, tick_skip=8, seconds_left=300, forfeit_spg_limit=None):
        # NOTE: Since game isn't reset to kickoff by default,
        # you need to keep this outside the main loop as well,
        # checking the done variable after each GoalTerminal
        super().__init__()
        self.tick_skip = tick_skip
        self.seconds_left = seconds_left
        self.timer = seconds_left
        self.overtime = False
        self.done = True
        self.initial_state = None
        self.forfeit_spg_limit = forfeit_spg_limit

    def reset(self, initial_state: GameState):
        if self.done:
            self.timer = self.seconds_left
            self.initial_state = initial_state
            self.overtime = False
            self.done = False

    def is_terminal(self, current_state: GameState) -> bool:
        differential = (current_state.blue_score - self.initial_state.blue_score) \
                       - (current_state.orange_score - self.initial_state.orange_score)
        if self.overtime:
            if differential != 0:
                self.done = True
            else:
                self.timer += self.tick_skip / 120
                self.done = False
        else:
            if self.timer <= 0 and current_state.ball.position[2] <= 100:
                # Can't detect ball on ground directly, should be an alright approximation
                if differential != 0:
                    self.done = True
                else:
                    self.overtime = True
                    self.done = False
            elif self.forfeit_spg_limit is not None \
                    and abs(differential) >= 3 \
                    and self.timer / abs(differential) < self.forfeit_spg_limit:
                self.done = True
            else:
                self.timer -= self.tick_skip / 120

        return self.done



"""
A terminal condition.
"""

from abc import ABC, abstractmethod
from rlgym.utils.gamestates import GameState


class TerminalCondition(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def reset(self, initial_state: GameState):
        """
        Function to be called each time the environment is reset.

        :param initial_state: The initial state of the reset environment.
        """
        raise NotImplementedError

    @abstractmethod
    def is_terminal(self, current_state: GameState) -> bool:
        """
        Function to determine if a game state is terminal. This will be called once per step, and must return either
        `True` or `False` if the current episode should be terminated at this state.

        :param current_state: The current state of the game.

        :return: Bool representing whether the current state is a terminal one.
        """
        raise NotImplementedError



import os
import pickle

import numpy as np
from rlgym_tools.replay_converter import convert_replay
from rlgym.utils.gamestates import GameState
from typing import List


def absolute_file_paths(directory: str):
    for dir_path, _, filenames in os.walk(directory):
        for f in filenames:
            yield os.path.abspath(os.path.join(dir_path, f))


def convert_replays(paths_to_each_replay: List[str], frame_skip: int = 150, verbose: int = 0) -> np.ndarray:
    """
    Converts replays into

    :param paths_to_each_replay: List of paths to each replay to be converted
    :param frame_skip: Every frame_skip frame from the replay will be converted
    :param verbose: When set to more than 0, it prints debug info
    :return: Numpy array of frames
    """
    states = []
    for replay in paths_to_each_replay:
        make_state_data(replay, states, frame_skip)
        if verbose > 0:
            print(replay, "done")

    return np.asarray(states)


def make_state_data(replay: str, states: list, frame_skip: int):
    """
    Process single replay

    :param replay: Path to the replay to process
    :param states: List of all the processed states
    :param frame_skip: Every frame_skip frame from the replay will be converted
    """
    replay_iterator = convert_replay(replay)
    for i, value in enumerate(replay_iterator):
        if i % frame_skip == frame_skip - 1:
            state, _ = value
            np_state = state_to_np_array(state)
            states.append(np_state)


def state_to_np_array(game_state: GameState) -> np.ndarray:
    """
    Gets only the useful information from game_state and puts it into numpy array

    :param game_state: GameState to convert
    :return: Numpy array of the most useful information from the game_state
    """
    whole_state = []
    ball = game_state.ball
    ball_state = np.concatenate((ball.position, ball.linear_velocity, ball.angular_velocity))

    whole_state.append(ball_state)
    for player in game_state.players:
        whole_state.append(np.concatenate((player.car_data.position,
                                           player.car_data.euler_angles(),
                                           player.car_data.linear_velocity,
                                           player.car_data.angular_velocity,
                                           np.asarray([player.boost_amount]))))
    return np.concatenate(whole_state)


def main():
    replay_names = list(absolute_file_paths("replays/3"))
    converted_states = convert_replays(replay_names, verbose=1)
    print(converted_states[10])
    with open("saved_gamestates3.gamestate", "wb") as f:
        pickle.dump(converted_states, f)


if __name__ == '__main__':
    main()



import numpy as np
from rlgym.utils import RewardFunction
from rlgym.utils.gamestates import PlayerData, GameState


class MultiplyRewards(RewardFunction):
    def __init__(self, *reward_functions: RewardFunction):
        super().__init__()
        self.reward_functions = reward_functions

    def reset(self, initial_state: GameState):
        for rew_func in self.reward_functions:
            rew_func.reset(initial_state)

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        return np.product([rf.get_reward(player, state, previous_action) for rf in self.reward_functions])

    def get_final_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        return np.product([rf.get_final_reward(player, state, previous_action) for rf in self.reward_functions])



from .air_dribble import air_dribble_config
from .dribble import dribble_config
from rlgym.envs import Match


def match_from_config(config, **additional_kwargs):
    for key, value in config.items():
        if key in additional_kwargs:
            print(f"Overwriting make argument {key}...")
        additional_kwargs[key] = value

    return Match(**additional_kwargs)



import logging
from collections import Counter
from typing import Union

import carball as cb
import numpy as np
from carball.analysis.analysis_manager import AnalysisManager
from carball.controls.controls import ControlsCreator

from rlgym.utils import math
from rlgym.utils.common_values import ORANGE_TEAM, BLUE_TEAM, BOOST_LOCATIONS
from rlgym.utils.gamestates import GameState, PhysicsObject, PlayerData

boost_locations = np.array(BOOST_LOCATIONS)  # Need ndarray for speed

_invert = np.array((-1, -1, 1))
_rot_correct = np.array((-1, 1, -1))


def convert_replay(replay: Union[str, AnalysisManager]):
    if isinstance(replay, str):
        replay = cb.analyze_replay_file(replay, logging_level=logging.CRITICAL)
    ControlsCreator().get_controls(replay.game)

    boost_timers = np.zeros(34)
    demo_timers = np.zeros(len(replay.game.players))

    blue_goals = 0
    orange_goals = 0
    goals = list(replay.game.goals)[::-1]
    touches = list(replay.protobuf_game.game_stats.hits)[::-1]
    demos = list(replay.game.demos)[::-1]

    match_goals = Counter()
    match_saves = Counter()
    match_shots = Counter()
    match_demos = Counter()
    match_boost_pickups = Counter()

    boost_amounts = {}
    last_locations = {}

    player_pos_pyr_vel_angvel_boost_controls = {  # Preload useful arrays so we can fetch by index later
        player.online_id: (
            player.data[["pos_x", "pos_y", "pos_z"]].values.astype(float),
            player.data[["rot_x", "rot_y", "rot_z"]].fillna(0).values.astype(float) * _rot_correct,
            player.data[["vel_x", "vel_y", "vel_z"]].fillna(0).values.astype(float) / 10,
            player.data[["ang_vel_x", "ang_vel_y", "ang_vel_z"]].fillna(0).values.astype(float) / 1000,
            player.data["boost"].fillna(0).astype(float) / 255,
            player.controls[["throttle", "steer", "pitch", "yaw", "roll",
                             "jump", "boost", "handbrake"]].fillna(0).values.astype(float),
        )
        for player in replay.game.players
    }

    ball_pos_pyr_vel_angvel = (
        replay.game.ball[["pos_x", "pos_y", "pos_z"]].values.astype(float),
        replay.game.ball[["rot_x", "rot_y", "rot_z"]].fillna(0).values.astype(float) * _rot_correct,
        replay.game.ball[["vel_x", "vel_y", "vel_z"]].fillna(0).values.astype(float) / 10,
        replay.game.ball[["ang_vel_x", "ang_vel_y", "ang_vel_z"]].fillna(0).values.astype(float) / 1000,
    )

    rallies = []
    for kf1, kf2 in zip(replay.game.kickoff_frames, replay.game.kickoff_frames[1:] + [replay.game.frames.index[-1]]):
        for goal in replay.game.goals:
            if kf1 < goal.frame_number < kf2:
                rallies.append((kf1, goal.frame_number))
                break
        else:  # No goal between kickoffs
            rallies.append((kf1, kf2))

    last_frame = 0
    for i, (frame, ball_row) in enumerate(replay.game.ball.iterrows()):
        for start, end in rallies:
            if start <= frame < end:
                # del rallies[0]
                break
        else:
            continue

        state = GameState()

        # game_type
        state.game_type = -1

        # blue_score/orange_score
        if len(goals) > 0 and goals[-1].frame_number <= frame:
            goal = goals.pop()
            match_goals[goal.player.online_id] += 1
            if goal.player_team == 0:
                blue_goals += 1
            else:
                orange_goals += 1
        state.blue_score = blue_goals
        state.orange_score = orange_goals

        # last_touch
        touched = set()
        while len(touches) > 0 and touches[-1].frame_number <= frame:
            touch = touches.pop()
            p_id = touch.player_id.id
            state.last_touch = p_id
            touched.add(p_id)
            if touch.save:
                match_saves[p_id] += 1
            if touch.shot:
                match_shots[p_id] += 1

        # demos for players
        demoed = set()
        while len(demos) > 0 and demos[-1]["frame_number"] <= frame:
            demo = demos.pop()
            attacker = demo["attacker"].online_id
            victim = demo["victim"].online_id
            match_demos[attacker] += 1
            demoed.add(victim)

        # players
        actions = []
        for n, player in enumerate(replay.game.players):
            player_data = PlayerData()
            if player.online_id in demoed:
                demo_timers[n] = 3

            player_data.car_id = player.online_id
            player_data.team_num = ORANGE_TEAM if player.team.is_orange else BLUE_TEAM
            player_data.match_goals = match_goals[player.online_id]
            player_data.match_saves = match_saves[player.online_id]
            player_data.match_shots = match_shots[player.online_id]
            player_data.match_demolishes = match_demos[player.online_id]
            player_data.boost_pickups = match_boost_pickups[player.online_id]
            player_data.is_demoed = demo_timers[n] > 0
            player_data.on_ground = None  # Undefined
            player_data.ball_touched = player.online_id in touched
            player_data.has_flip = None  # Undefined, TODO use jump_active, double_jump_active and dodge_active?
            pos, pyr, vel, ang_vel, boost, controls = (v[i] for v in
                                                       player_pos_pyr_vel_angvel_boost_controls[player.online_id])

            player_data.boost_amount = boost
            if np.isnan(pos).any():
                pos = last_locations[player.online_id]
            else:
                last_locations[player.online_id] = pos
            player_data.car_data = PhysicsObject(
                position=pos,
                quaternion=math.rotation_to_quaternion(math.euler_to_rotation(pyr)),
                linear_velocity=vel,
                angular_velocity=ang_vel
            )
            player_data.inverted_car_data = PhysicsObject(
                position=pos * _invert,
                quaternion=math.rotation_to_quaternion((math.euler_to_rotation(pyr).T * _invert).T),
                linear_velocity=vel * _invert,
                angular_velocity=ang_vel * _invert
            )

            old_boost = boost_amounts.get(player.online_id, float("inf"))
            boost_change = boost - old_boost
            boost_amounts[player.online_id] = player_data.boost_amount
            if boost_change > 0 and not (old_boost == 0 and boost == 85 / 255):  # Ignore boost gains on spawn
                closest_boost = np.linalg.norm(boost_locations - pos, axis=-1).argmin()
                if boost_locations[closest_boost][1] > 72:
                    boost_timers[closest_boost] = 10
                else:
                    boost_timers[closest_boost] = 4
                match_boost_pickups[player.online_id] += 1

            state.players.append(player_data)

            actions.append(controls)

        # ball
        pos, pyr, vel, ang_vel = (v[i] for v in ball_pos_pyr_vel_angvel)
        if np.isnan(pos).any():
            continue  # Goal scored, go next
        state.ball = PhysicsObject(
            position=pos,
            quaternion=math.rotation_to_quaternion(math.euler_to_rotation(pyr)),
            linear_velocity=vel,
            angular_velocity=ang_vel
        )

        # inverted_ball
        state.inverted_ball = PhysicsObject(
            position=pos * _invert,
            quaternion=math.rotation_to_quaternion((math.euler_to_rotation(pyr).T * _invert).T),
            linear_velocity=vel * _invert,
            angular_velocity=ang_vel * _invert
        )

        # boost_pads
        state.boost_pads = (boost_timers == 0) * 1

        # inverted_boost_pads
        state.inverted_boost_pads = state.boost_pads[::-1]

        d_time = (frame - last_frame) / 30  # Maybe use time delta from replay instead?
        boost_timers -= d_time  # Should this be before or after values are set?
        demo_timers -= d_time
        boost_timers[boost_timers < 0] = 0
        demo_timers[demo_timers < 0] = 0
        last_frame = frame

        state.players.sort(key=lambda p: (p.team_num, p.car_id))

        yield state, actions



import math
import numpy as np
from typing import Any, List
from rlgym.utils import common_values
from rlgym.utils.gamestates import PlayerData, GameState
from rlgym.utils.obs_builders import ObsBuilder


class DefaultObs(ObsBuilder):
    def __init__(self, pos_coef=1/2300, ang_coef=1/math.pi, lin_vel_coef=1/2300, ang_vel_coef=1/math.pi):
        """
        :param pos_coef: Position normalization coefficient
        :param ang_coef: Rotation angle normalization coefficient
        :param lin_vel_coef: Linear velocity normalization coefficient
        :param ang_vel_coef: Angular velocity normalization coefficient
        """
        super().__init__()
        self.POS_COEF = pos_coef
        self.ANG_COEF = ang_coef
        self.LIN_VEL_COEF = lin_vel_coef
        self.ANG_VEL_COEF = ang_vel_coef

    def reset(self, initial_state: GameState):
        pass

    def build_obs(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> Any:
        if player.team_num == common_values.ORANGE_TEAM:
            inverted = True
            ball = state.inverted_ball
            pads = state.inverted_boost_pads
        else:
            inverted = False
            ball = state.ball
            pads = state.boost_pads

        obs = [ball.position * self.POS_COEF,
               ball.linear_velocity * self.LIN_VEL_COEF,
               ball.angular_velocity * self.ANG_VEL_COEF,
               previous_action,
               pads]

        self._add_player_to_obs(obs, player, inverted)

        allies = []
        enemies = []

        for other in state.players:
            if other.car_id == player.car_id:
                continue

            if other.team_num == player.team_num:
                team_obs = allies
            else:
                team_obs = enemies

            self._add_player_to_obs(team_obs, other, inverted)

        obs.extend(allies)
        obs.extend(enemies)
        return np.concatenate(obs)

    def _add_player_to_obs(self, obs: List, player: PlayerData, inverted: bool):
        if inverted:
            player_car = player.inverted_car_data
        else:
            player_car = player.car_data

        obs.extend([
            player_car.position * self.POS_COEF,
            player_car.forward(),
            player_car.up(),
            player_car.linear_velocity * self.LIN_VEL_COEF,
            player_car.angular_velocity * self.ANG_VEL_COEF,
            [player.boost_amount,
             int(player.on_ground),
             int(player.has_flip),
             int(player.is_demoed)]])

        return player_car



# From https://github.com/RLBot/RLBot/blob/master/src/main/python/rlbot/version.py
# Store the version here so:
# 1) we don't load dependencies by storing it in __init__.py
# 2) we can import it in setup.py for the same reason
# 3) we can import it into your module module
# https://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package

__version__ = '1.1.0'

release_notes = {
    'beta':
    """
    - This version contains numerous untested and potentially breaking changes. Run at your own risk.
    """,
    '1.1.0': """
    - Added ActionParsers, which allow you to define custom action spaces - Lolaapk3
    - Fixed quaternion to euler angle conversion - Darxeal
    - Improved DefaultObs, it now has all the state information
    - Added optional parameter to return initial info object on env.reset()
    - Fixed rand_vec3 so the magnitude is actually random
    """,
    '1.0.2': """
    - Fixed state setting issues (boost setting and precision)
    """,
    '1.0.1': """
    - Minor fix for RandomState
    """,
    '1.0.0': """
    - Added state setting, you can now specify the initial state of episodes via a StateSetter object - aiTan
    - Added launch_preference to make()
    - Fixed Epic launch bug introduced with the Season 4 update
    - Communication code refactored to increase performance with low tick_skip
    - Fixed boost pad reset bug on episode reset
    - Added detection mechanism for ghost ball bug
    - Decreased required access rights by the MultiInjector
    """,
    '0.6.0': """
    - Moved wrappers and replay converter to rlgym-tools package
    - Added the optional ability to forcefully page the spawned Rocket League instances upon creation - 416c616e
    - The path_to_rl param is no longer required for use_injector
    """,
    '0.5.0': """
    - Removed string based configurations in rlgym.make(), everything is passed by kwargs now - Soren
    - Added StableBaselines3 compatibility - Rolv
    - Refactored and expanded reward functions - Rolv
    - Added replay converter - Rolv
    - Fixed TouchBallReward bug - Kevin

    NOTE: Some of these new tools (wrappers, replay converter, etc) will be moved to a different package in the next release
    """,
    '0.4.1': """
    - Updated euler angles to match rlbot [pitch, yaw, roll] and added accessor functions
    - Bugfix: player.is_alive renamed to is_demoed
    - Added common rewards - Rolv
    - Added a reward combiner - Chainso
    - Added missing kickoff spawn
    - Fixed 2v2 and 3v3 action delivery
    - Fixed issue in 2v2 and 3v3 where blue bots would disappear over time
    - Added multi injector
    """,
    '0.4.0': """
    - Major API refactor
    - Added boostpad boolean array to GameState
    - RLGym is now baselines compatible
    - Added improved Default configuration
    """,
    '0.3.0': """
    - Pass initial state to env components on gym.reset()
    - Pass prev action to reward fn
    - info returned from gym.step() is now a Dict
    - Fixed obs size bug in RhobotObs
    """,
    '0.2.0': """
    - Switched from custom_args dict to keyword args in rlgym.make()
    """,
    '0.1.4': """
    - Gym now inherits openai.gym.Env - Chainso
    - Fixed bug where rlgym crashed trying to parse the state string in some scenarios
    """,
    '0.1.3': """
    - Renamed PhysicsObject.orientation to euler_angles
    - euler_angles is now a function so they aren't calculated if not needed
    - Added rotation_mtx, forward, right and up vectors to PhysicsObject (only usable for the car)
    - Removed conversion to np array for rewards and observations
    - Fixed bug in RhobotObs
    """,
    '0.1.2': """
    - Obs are now numpy arrays
    - Added types to core classes to make customization easier
    - Fixed bakkesmod bug
    """,
    '0.1.1': """
    Added missing scipy dependency
    """,
    '0.1.0': """
    Initial Release
    """
}


def get_current_release_notes():
    if __version__ in release_notes:
        return release_notes[__version__]
    return ''


def print_current_release_notes():
    print(f"Version {__version__}")
    print(get_current_release_notes())
    print("")


class Message(object):
    RLGYM_HEADER_END_TOKEN                               = [13771, 83712, 83770]
    RLGYM_BODY_END_TOKEN                                 = [82772, 83273, 83774]
    RLGYM_NULL_MESSAGE_HEADER                            = [83373, 83734, 83775]
    RLGYM_NULL_MESSAGE_BODY                              = [83744, 83774, 83876]
    RLGYM_CONFIG_MESSAGE_HEADER                          = [83775, 53776, 83727]
    RLGYM_STATE_MESSAGE_HEADER                           = [63776, 83777, 83778]
    RLGYM_AGENT_ACTION_MESSAGE_HEADER                    = [87777, 83778, 83779]
    RLGYM_RESET_GAME_STATE_MESSAGE_HEADER                = [83878, 83779, 83780]
    RLGYM_AGENT_ACTION_IMMEDIATE_RESPONSE_MESSAGE_HEADER = [83799, 83780, 83781]
    RLGYM_REQUEST_LAST_BOT_INPUT_MESSAGE_HEADER          = [83781, 83781, 83682]
    RLGYM_LAST_BOT_INPUT_MESSAGE_HEADER                  = [11781, 83782, 83983]
    RLGYM_RESET_TO_SPECIFIC_GAME_STATE_MESSAGE_HEADER    = [12782, 83783, 80784]

    @staticmethod
    def deserialize_header(message_floats):
        assert type(message_floats) in (list, tuple), "!ATTEMPTED TO DECODE MESSAGE HEADER FROM NON-LIST TYPE! []".format(type(message_floats))
        m = message_floats
        header_end = Message._find_first(m, Message.RLGYM_HEADER_END_TOKEN)
        return m[:header_end]

    @staticmethod
    def _find_first(lst, target):
        n = len(target)
        for i in range(len(lst)):
            if lst[i] == target[0] and lst[i:i+n] == target:
                return i
        return None

    def __init__(self, header=None, body=None):
        if header is None:
            header = Message.RLGYM_NULL_MESSAGE_HEADER
        if body is None:
            body = Message.RLGYM_NULL_MESSAGE_BODY

        self.body = body
        self.header = header

    def serialize(self):
        return self.header + Message.RLGYM_HEADER_END_TOKEN + self.body + Message.RLGYM_BODY_END_TOKEN

    def deserialize(self, message_floats):
        assert type(message_floats) == list, "!ATTEMPTED TO DECODE MESSAGE HEADER FROM NON-LIST TYPE! []".format(type(message_floats))
        m = message_floats

        header_end = Message._find_first(m, Message.RLGYM_HEADER_END_TOKEN)
        header = m[:header_end]

        start = Message._find_first(m, Message.RLGYM_HEADER_END_TOKEN) + len(Message.RLGYM_HEADER_END_TOKEN)
        end = Message._find_first(m,Message.RLGYM_BODY_END_TOKEN)
        body = m[start:end]

        self.body = body
        self.header = header

